from mobspy import *

if __name__ == '__main__':

    """
        This is the Tree model from the paper
        We have a population of Trees
        The Trees can die, age, have different colors and be in two different forests
        The colors can change randomly from time to time
        All old Trees can reproduce, but the Three is born green and young 
    """
    Ager, Mortal, Colored, Location = BaseSpecies(4)
    Colored.green, Colored.yellow, Colored.brown
    Location.dense, Location.sparse
    Ager.young >> Ager.old[1 / 10 / u.year]
    Mortal >> Zero[lambda r1: 0.1/ u.year if r1.old else 0]
    Tree = Ager * Colored * Mortal * Location

    # replication
    Tree.old >> Tree + Tree.green.young[0.1/u.year]

    # competition
    Tree.dense.old + Tree.dense.young >> Tree.dense.old \
        [1e-10 * u.decimeter / u.year]

    # color cycling
    colors = ['green', 'yellow', 'brown']
    for color, next_color in zip(colors, colors[1:] + colors[:1]):
        Tree.c(color) >> Tree.c(next_color)[10/u.year]

    # initial conditions
    Tree.dense(50), Tree.dense.old(50), Tree.sparse(50), Tree.sparse.old(50)
    MySim = Simulation(Tree)
    print(MySim.compile())
    exit()
    MySim.simulation_method = 'stochastic'
    MySim.save_data = False
    MySim.plot_data = False
    MySim.duration = 100*u.years
    MySim.unit_x = 'year'
    MySim.run()
    MySim.plot_stochastic(Tree.dense, Tree.sparse, Tree.brown)


from copy import deepcopy
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.plot_params.example_plot_reader as epr


def query_plot_data(species, data):
    """
        Performs a query of the plot data, when one wishes to plot species with characteristics
        It creates a new data structure with the results from the query added to it for the plotting structure

        :param species: (str) Species name in string format
        :param data: (dict) Data in MobsPy dictionary format
    """
    new_data = deepcopy(data)

    species_to_plot = set()
    for time_series in data.ts_data:
        for key in time_series.keys():
            if key in species:
                species_to_plot.add(key)

    for spe in species:
        if spe in species_to_plot:
            continue

        for i, ts in enumerate(new_data.ts_data):
            new_data.time_series_number = i
            ts[spe] = new_data[spe]
            species_to_plot.add(spe)

    return species_to_plot, new_data


def check_plot_parameters(species, plot_params):
    """
        Performs a check of the plot_parameters given. To see if the parameters are correctly named

        :param species: (str) Species in str format
        :param plot_params: (dict) Plot parameter dictionary
    """
    dictionary = epr.get_example_plot_parameters()

    if 'Time' in plot_params:
        simlog.error('Time must not be a plot parameter name')

    for spe in species:
        if spe in dictionary.keys():
            simlog.error(f'Plotting is impossible, species {spe} is a parameter name')

    for key in plot_params:
        if key not in dictionary and key not in species:
            simlog.warning(f'Parameter {key} not supported')






"""
    This model stores function used by the meta_class.py module
"""
import copy
import itertools
import mobspy.simulation_logging.log_scripts as simlog


def combine_references(species1, species2):
    """
        Combine the sets of references of two species

        :param species1: (Meta-species object)
        :param species2: (Meta-species object)
    """
    return species1.get_references().union(species2.get_references())


def check_orthogonality_between_references(references):
    """
        Check if meta-species objects inside a reference do not have characteristics in common
        The sets of characteristics directly added to species must be independent

        :param references: (set) set of meta-species objects to check for independence
        :raise simlog.error: raises error if there are repeated characteristics in different meta-species
    """
    for i, reference1 in enumerate(references):
        for j, reference2 in enumerate(references):

            if i == j:
                continue

            if len(reference1.get_characteristics().intersection(reference2.get_characteristics())) != 0:
                simlog.error(f'The same characteristic can only be shared through inheritance. ' +
                             f'There are two characteristics directly added to two meta-species \n'
                             f'Repetition in: {reference1}, {reference2}'
                             f'Characteristics: {reference1.get_characteristics()}, {reference2.get_characteristics()}')


def complete_characteristics_with_first_values(spe_object, characteristics, characteristics_to_object):
    """
        This creates a string with the species object name and the set of all first characteristics added
        to it's base species used to construct it
        It allows us to search for the proper string in the model using the result of this function
        It's used to set the quantities using the call method in Species and Reacting_Species

        :param spe_object: (meta-species object) meta-species to generate the species states with the first values
        :param characteristics: (str) if there are characteristics to use instead of the default
        :param characteristics_to_object: (dict) dictionary with characteristics as keys and the meta-species which
        they have been added to directly as value (no inheritance or New)

        :return: Set with the species name and characteristics
    """
    if characteristics == 'std$':
        characteristics = set()

    vector_elements = {}
    for cha in characteristics:
        vector = characteristics_to_object[cha]
        if vector in vector_elements:
            simlog.error('The assignment refers to multiple strings')
        else:
            vector_elements[vector] = True

    first_characteristics = set()
    for reference in spe_object.get_references() - set(vector_elements.keys()):
        if reference.first_characteristic:
            first_characteristics.add(reference.first_characteristic)

    return {spe_object.get_name()}.union(first_characteristics).union(characteristics)


def unite_characteristics(species):
    """
        This function unites the characteristics of all the given species

        :param species: (list of species or List_Species object)
    """
    characteristics = set()

    if species is not None:
        for spe in species:
            characteristics = characteristics.union(spe.get_characteristics())

    return characteristics


def create_orthogonal_vector_structure(species):
    """
        This creates the independent state-structure for the model
        It is just a dictionary where the keys are characteristics and the values are meta-species objects that have
        been directly added to that object (no inheritance or New)
        It simplifies the code by allowing to easily keep track of the 'axis' of each characteristic. Allowing
        for easy transformation on the products and others

        :param species: (meta-species objects) - meta-species objects used in a model

        :return ref_characteristics_to_object: (dict) a dictionary where the keys are characteristics and the
        values are meta-species objects that have been directly added to that object
    """
    ref_characteristics_to_object = {}
    for spe in species:
        for prop in spe.get_references():
            for cha in prop.get_characteristics():

                if cha not in ref_characteristics_to_object:
                    ref_characteristics_to_object[cha] = prop
                elif ref_characteristics_to_object[cha] == prop:
                    pass
                else:
                    simlog.error(f'The same characteristic can only be shared through inheritance. ' +
                                 f'There are two characteristics directly added to two meta-species \n'
                                f'Repetition in: {spe}, {ref_characteristics_to_object[cha] } \n'
                                f'Characteristics: {spe.get_characteristics()}, '
                                 f'{ref_characteristics_to_object[cha].get_characteristics()} \n')

    return ref_characteristics_to_object


if __name__ == '__main__':
    pass



from mobspy import *

"""
    This example model was designed to talk about looping through species
    Since there are two ways, we leave it for the user to pick the best suited one
"""

"""
    In this model we loop through species in two different ways
"""


def first_loop():
    # Define a base species
    # And an inheritor through the New operator
    DNAStrains = BaseSpecies()
    DNACounter = New(DNAStrains, 1)

    # Here we loop through the species by assigning different characteristics
    for i in range(10):
        DNACounter.c(f"dna_{i}")
    print(DNACounter.get_characteristics())


"""
    The New Function can also receive a name instead of a number
    The name allows one to name the species with the string instead of it's variable
"""


def second_loop():
    DNAStrains = BaseSpecies()

    # In this model we use the New operator to create different meta-species
    # And store them in the DNAList variable using the | operator
    # Here since there are no variables to store each meta-species
    # the new function receives a string which will be the name of the meta-species
    # When it receives a string it only returns one species with the string as it's name
    DNAList = None
    for i in range(10):
        DummyDNA = New(DNAStrains,f"DNA_{i}")
        if DNAList is None:
            DNAList = DummyDNA
        else:
            DNAList = DummyDNA | DNAList

    # One can then loop through the DNAList using an iterator like a for loop
    for DNA in DNAList:
        print(DNA)


if __name__ == '__main__':
    first_loop()
    second_loop()




import json
import os.path
import math
import mobspy.plot_scripts.statistics_calculations as spd
import mobspy.simulation_logging.log_scripts as simlog

import numpy as np
import pickle as pkl
from pathlib import Path
import inspect
import matplotlib.pyplot as plt


####################### PRACTICAL FUNCTIONS
class Color_cycle:
    """
        This class is responsible for cycling through the different colors for different curves

        :param index: (int) Current position in the color cycle
        :param color_list: (list) = list of available colors
    """

    def __init__(self):
        self.index = 0
        self.color_list = ['b', 'g', 'r', 'c', 'm', 'y', 'k']

    def __call__(self, n):
        """
            Call updates the index and returns the next color in the list
            Normally n is set to 1

            :param n: (int) number of positions to skip
            :return: color (str) = color name for pyplot
        """
        self.index = (self.index + n) % len(self.color_list)
        return self.color_list[self.index]


def find_species_time_series(spe, data):
    """
        There can be different time-series in MobsPy data (even experimental data, as long as it is in MobPy format)
        This function finds all the time-series the species is present in and returns it for looping thorugh all of
        them
        This function is implemented to allow for the comparison of models with experimental data or other models

        :param spe: (str) Species name
        :pram  data: (dict) Data in MobsPy format
    """
    for time_series in data:
        if spe in time_series:
            yield time_series


def get_total_figure_number(axis_matrix):
    """
        Gets the number of figures from an axis_matrix generated by pyplot. Used by the hash to place configs in the
        correct part of the axis_matrix

        :param axis_matrix: (1D or 2D numpy array) Array returned by pyplot once multiple figures are introduced into
        a single plot

        :return: total_figure_number (int) Total number of figures
    """
    # Get the number of figures, using the axis_matrix
    try:
        total_figure_number = axis_matrix.shape[0] * axis_matrix.shape[1]
    except IndexError:
        total_figure_number = axis_matrix.shape[0]
    return total_figure_number


# Hash for converting linear figure number into index
def figure_hash(current_figure, axis_matrix):
    """
        This function allows one to acess the figure grid with a linear input
        For instance one can access a 2x2 grid using 0, 1, 2, 3
        0 becomes 0,0
        1 becomes 1,0
        2 becomes 0,1
        3 becomes 1,1

        :param current_figure: (int) linear number of the figure
        :param axis_matrix: (1D or 2D numpy array) a list with all the created axis on the multiple figure subplot

        :return: the correct axis based on the number provided
    """

    # Get the number of lines
    max_lines = len(axis_matrix)
    total_figure_number = get_total_figure_number(axis_matrix)

    # Correction for index purposes
    current_figure = current_figure

    col = math.floor(current_figure / max_lines)

    if total_figure_number <= max_lines:
        return axis_matrix[int(current_figure % max_lines)]
    else:
        return axis_matrix[int(current_figure % max_lines), int(col)]


# Hash to convert total figure number into grid
def figure_hash_creation(total_figure_number, max_lines=None):
    """

        This is hash used to create the figure grid automatically according to the number of figures desired by the
        user and the maximum number of lines

        For instance 4 figures with 2 as max_lines creates a 2x2 figure grid automatically

        :param total_figure_number: (int) Number of total figures to create
        :param max_lines: (int) Maximum number of lines in the grid

        :return: fig and axs = Figures grid will all the respective axis
    """

    # Default value if nothing is set up
    if max_lines is None:
        max_lines = 2

    if total_figure_number <= max_lines:
        fig, axs = plt.subplots(total_figure_number)
    else:
        column_number = int((total_figure_number + total_figure_number % max_lines) / max_lines)
        fig, axs = plt.subplots(max_lines, column_number)

    if total_figure_number == 1:
        axs = np.array([axs])

    return fig, axs


def find_parameter(params, key, index=None):
    """

        This is the hearth of the plotting structure, this functions allows one to simply set multiple characteristics
        for only one figure

        The priority for parameter search is plots => figures => global, with plot overriding others and so on

        If a parameter is defined globally it will be applied to all figures, if it defined inside a figure element
        it will only apply to that figure, if it is defined in a plot element it will only apply to that curve

        Check the readme or the tutorials for more details on the plotting structure. It is simple and versatile

        :param params: (dict) Plot parameters from python dictionary (after json conversion)
        :param key: (str) Key necessary to access the parameters
        :param index: (int) None for global search, one index for figure search, and two for figure curve search

        :return: The parameter if found, and None if nothing is found
    """

    # No index is given, look global
    if index is None:
        try:
            return params[key]
        except (KeyError, IndexError):
            return None
    # Search a parameter
    # If local return local, otherwise return global
    # If not found return nothing
    elif type(index) == int:

        try:
            return params['figures'][index][key]
        except (KeyError, IndexError):

            try:
                return params[key]
            except (KeyError, IndexError):
                return None
    # If two indexes are given
    # Look inside the plot
    # Than figure
    # Than global
    elif type(index) == tuple:

        try:
            return params['figures'][index[0]]['plots'][index[1]][key]

        except (KeyError, IndexError):

            try:
                return params['figures'][index[0]][key]

            except (KeyError, IndexError):

                try:
                    return params[key]
                except (KeyError, IndexError):
                    return None


####################### PLOTING FUNCTIONS

def plot_curves(data, axs, figure_index, plot_params):
    """
        This function plots the programmed curves in the assigned figure

        :param axs: (pyplot axe) axs to plot the data in
        :param data: (dict) data given in MobsPy format results['data']
        :param figure_index: (int) index of the current figure to plot curves in
        :param plot_params: parameters for plotting
    """

    # Get the plot number from the list of plots
    try:
        plot_number = len(find_parameter(plot_params, 'plots', figure_index))
    except TypeError:
        # No figures plot only the default
        plot_number = 1
    if plot_number == 0:
        plot_number = 1

    # For all plots in the figure
    # Set all parameters and plot
    legend_flag = False
    for plot_index in range(plot_number):

        # Get the species from plot parameters
        if find_parameter(plot_params, key='species_to_plot', index=(figure_index, plot_index)) is not None:
            species = find_parameter(plot_params, key='species_to_plot', index=(figure_index, plot_index))
        else:
            simlog.error('No species found for plotting in one of the curves or figures')

        # Get the time series to plot
        if find_parameter(plot_params, key='time_series', index=(figure_index, plot_index)) is not None:
            time_series = find_parameter(plot_params, key='time_series', index=(figure_index, plot_index))
            if type(time_series) == int:
                time_series = [time_series]
        else:
            time_series = list(range(len(data)))

        for spe in species:

            color_index = 0

            # Get the parameters assigned to the species, if not assign empty for None returns
            if find_parameter(plot_params, key=spe, index=(figure_index, plot_index)) is not None:
                species_characteristics = find_parameter(plot_params, key=spe, index=(figure_index, plot_index))
            else:
                species_characteristics = {}

            # Now we search the species characteristics for parameters
            if find_parameter(species_characteristics, key='color') is not None:
                curve_color = find_parameter(species_characteristics, key='color')
            else:
                color_index = (color_index + 1) % len(Color_cycle().color_list)
                curve_color = Color_cycle()(color_index)

            if find_parameter(species_characteristics, key='linestyle') is not None:
                linestyle = find_parameter(species_characteristics, key='linestyle')
            else:
                linestyle = '-'

            if find_parameter(species_characteristics, key='linewidth') is not None:
                linewidth = find_parameter(species_characteristics, key='linewidth')
            else:
                linewidth = None

            if find_parameter(species_characteristics, key='label') is not None:
                label = find_parameter(species_characteristics, key='label')
                legend_flag = True
            else:
                label = None

            for ts in time_series:
                ts = data[ts]

                try:
                    if find_parameter(plot_params, key='fill_between', index=(figure_index, plot_index)) is not None \
                            and find_parameter(plot_params, key='fill_between', index=(figure_index, plot_index)):
                        try:
                            axs.fill_between(ts['Time'], ts[spe][0], ts[spe][1], color=curve_color,
                                             label=label)
                        except IndexError:
                            simlog.error('Fill_between must only have two or less runs referring to it')
                    else:
                        axs.plot(ts['Time'], ts[spe], color=curve_color,
                                 linestyle=linestyle, linewidth=linewidth, label=label)
                        label = None
                except KeyError:
                    pass

    if legend_flag:
        if find_parameter(plot_params, key='frameon', index=figure_index) is not None:
            frameon = find_parameter(plot_params, key='frameon', index=figure_index)
        else:
            frameon = True
        axs.legend(frameon=frameon)


def set_figure_characteristics(axis_matrix, plot_params):
    """
        Sets the characteristics for each figure

        :param axis_matrix: (axis from pyplot subplot) Array of all axis in the grid
        :param plot_params: (dict) Plot parameters received
    """
    total_figure_number = get_total_figure_number(axis_matrix)
    # Loop through all axis
    for i in range(total_figure_number):

        if find_parameter(plot_params, 'xlim', i) is not None:
            figure_hash(i, axis_matrix).set_xlim(find_parameter(plot_params, 'xlim', i))

        if find_parameter(plot_params, 'ylim', i) is not None:
            figure_hash(i, axis_matrix).set_ylim(find_parameter(plot_params, 'ylim'))

        if find_parameter(plot_params, 'logscale', i) is not None and 'X' in find_parameter(plot_params, 'logscale', i):
            figure_hash(i, axis_matrix).set_xscale('log')

        if find_parameter(plot_params, 'logscale', i) is not None and 'Y' in find_parameter(plot_params, 'logscale', i):
            figure_hash(i, axis_matrix).set_yscale('log')

        if find_parameter(plot_params, 'title', i) is not None:
            figure_hash(i, axis_matrix).set_title(find_parameter(plot_params, 'title', i))

        if find_parameter(plot_params, 'xlabel', i) is not None:
            figure_hash(i, axis_matrix).set_xlabel(find_parameter(plot_params, 'xlabel', i))

        if find_parameter(plot_params, 'ylabel', i) is not None:
            figure_hash(i, axis_matrix).set_ylabel(find_parameter(plot_params, 'ylabel', i))

        if find_parameter(plot_params, 'annotations', i) is not None:
            for annotations in find_parameter(plot_params, 'annotations', i):
                figure_hash(i, axis_matrix).text(annotations[0], annotations[1], annotations[2])


def set_global_parameters(fig, plot_params):
    """
        Sets the characteristics the plot window

        :param fig: (fig from pyplot subplot) Array of all axis in the grid
        :param plot_params: (dict) Plot parameters received
    """
    if find_parameter(plot_params, 'pad') is not None:
        fig.tight_layout(pad=find_parameter(plot_params, 'pad'))

    if find_parameter(plot_params, 'figsize') is not None:
        fig.set_size_inches(plot_params['figsize'][0], plot_params['figsize'][1])

    if find_parameter(plot_params, 'dpi') is not None:
        fig.set_dpi(plot_params['dpi'])


def plot_data(data, plot_params):
    """
        This function plots the simulation results according to the specifications

        :param data: (dict) Data in MobsPy format
        :param plot_params: (dict) Plot parameters received
    """

    # Get the figure number from the list of figures
    # Add it to parameters
    try:
        figure_number = len(find_parameter(plot_params, 'figures'))
    except TypeError:
        # No figures plot only the default
        figure_number = 1
    if figure_number == 0:
        figure_number = 1

    # Create figures grid, max_lines is 2 as default and the value is defined in the function
    fig, axis_matrix = figure_hash_creation(figure_number, max_lines=find_parameter(plot_params, 'max_lines'))

    # Set parameters common to all figures, like padding and figure size.
    set_global_parameters(fig, plot_params)

    # Global call for setting global parameters
    set_figure_characteristics(axis_matrix, plot_params)

    # Now we plot
    for figure_index in range(figure_number):
        plot_curves(data, figure_hash(figure_index, axis_matrix), figure_index, plot_params)

    # Real default case
    plt.show()


if __name__ == "__main__":
    pass



import mobspy.simulation_logging.log_scripts as simlog
import inspect
import mobspy.modules.species_string_generator as ssg
from pint import Quantity

"""
    This model implements the logic comparison structure for events and conditional durations.
    The syntax follows:( MetaSpeciesObject Operator Value ) ex: (A > 5)
    For clause combination one has - & and | or.
    All clauses must be inside parenthesis when combined due to Python operator precedence.
"""


class SpeciesComparator:
    """
        This class implements the comparisons necessary for events and conditional durations for Species
        ex (A > 5)

        :param _simulation_context: (Simulation) current simulation under context
    """

    @classmethod
    def check_parenthesis(cls, code_line, line_number, pos, symbol, number_of_comp):
        """
            This function compiles part of a logic expression for meta-species. It checks to see if an operator '<' or
            '>' is properly written inside parenthesis. It only applies if more than one operator is present in the
            code line

        :param code_line: (str) line of code to be compiled
        :param line_number: (str) number of the code line currently being compiled
        :param pos: (int) position of the '>' or '<' operator
        :param symbol: '(' or ')', indicates in which direction the string analysis should proceded
        :param number_of_comp: for distinction the default case where only one operator is present

        :raise simlog.error: if the code line is not properly written isolating the clauses with parenthesis

        :return: (bool) false if the line compiles, raises a error if it does not
        """
        i = pos
        condition_not_satisfied = True
        while condition_not_satisfied:
            try:
                if symbol == ')':
                    i += 1
                elif symbol == '(':
                    i -= 1

                # Check if finished
                if i == 0 or i == len(code_line):
                    if number_of_comp > 1:
                        simlog.error(f'At: {code_line} \n' + f'Line number: {line_number} \n' +
                                     'All clauses must be individually isolated by parenthesis '
                                     '- Ex: (A <= 0) & (B <= 0) \n')
                    else:
                        return False

                char = code_line[i]
                if char == '<' or char == '>' or char == '|' or char == '&':
                    simlog.error(f'At: {code_line} \n' + f'Line number: {line_number} \n' +
                                 'All clauses must be isolated by parenthesis - Ex: (A <= 0) & (B <= 0) \n')
                elif char == ')' and number_of_comp > 1:
                    if symbol == ')':
                        condition_not_satisfied = False
                elif char == '(' and number_of_comp > 1:
                    if symbol == '(':
                        condition_not_satisfied = False
            except IndexError:
                simlog.error(f'Error Compiling the following line {line_number}: {code_line}')
        return condition_not_satisfied

    @classmethod
    def compile_code_line(cls):
        """
            Compiles the code line executing a logical operation

            :raise simlog.error: if the code line is not properly written isolating the clauses with parenthesis
        """
        code_line = inspect.stack()[2].code_context[0][:-1]
        line_number = inspect.stack()[2].lineno
        if 'and' in code_line or 'or' in code_line:
            simlog.error(f'At: {code_line} \n' + f'Line number: {line_number} \n' +
                         'Event notation did not compile, please use & for \'and\' and  | for \'or\' \n'
                         'Please also put the clauses under parentheses: example (A <= 0) & (B <= 0)')

        temp_code_line = code_line.replace(' ', '')
        multiplication_position = [pos for pos, char in enumerate(temp_code_line) if char == '*']
        for pos in multiplication_position:
            if not (temp_code_line[pos - 1].isnumeric() or temp_code_line[pos + 1].isnumeric()):
                simlog.error(f'At: {code_line} \n' + f'Line number: {line_number} \n' +
                             'Multiplication between meta-species under comparison context' +
                             ' not yet supported by MobsPy - \n.')

        number_of_comp = code_line.count('<') + code_line.count('>')
        comparison_position = [pos for pos, char in enumerate(code_line) if char == '<' or char == '>']
        for pos in comparison_position:
            for symbol in ['(', ')']:
                assert not cls.check_parenthesis(code_line, line_number, pos, symbol, number_of_comp)

    def __init__(self):
        self._simulation_context = None

    def add_operation_and_number(self, symbol, number):
        """
            Creates a MetaSpeciesLogicResolver from the comparison of a meta-species with a value or another
            meta-species

            :param symbol: Comparison symbol '>=', '<=', '>' or '<'
            :param number: (int) if compared to a value, (Species or ReactingSpecies) if compared to the objects
        """
        number = self.reformat_number_and_species(number)
        if type(number) == list:
            operation = [self.logical_add_species(self)] + [symbol] + number
        else:
            operation = [self.logical_add_species(self)] + [symbol] + [number]
        logic_re_object = MetaSpeciesLogicResolver(operation, self._simulation_context)
        return logic_re_object

    def reformat_number_and_species(self, number):
        """
            Discovers if the number is a Species, ReactingSpecies or integer and prepares the output accordingly

            :param number: (int) if compared to a value, (Species or ReactingSpecies) if compared to the objects
        """
        if isinstance(number, SpeciesComparator):
            if number.is_species():
                number = self.logical_add_species(number)
            else:
                number = self.logical_add_reacting_species(number)
        return number

    @classmethod
    def logical_add_species(cls, species):
        """
            Adds a species object to an event trigger operation

            :param species: (Species) species object
        """

        return {'object': species, 'characteristics': set()}

    @classmethod
    def logical_add_reacting_species(cls, react_spe):
        """
            Adds a reacting species object to an event trigger operation by summing over all the indicated
            characteristics. It also accepts sums of species multiplied by integers

            :param react_spe: (ReactingSpecies) ReactingSpecies object
        """
        operation = []

        react_spe.check_context()
        for i, react_dict in enumerate(react_spe.list_of_reactants):
            if i > 0:
                dl = ['+']
            else:
                dl = []
            dl = dl + [react_dict['stoichiometry'], '*']
            dl = dl + [{'object': react_dict['object'], 'characteristics': react_dict['characteristics']}]
            operation = operation + dl
        return dl

    def __lt__(self, number):
        self.compile_code_line()
        return self.add_operation_and_number('<', number)

    def __le__(self, number):
        self.compile_code_line()
        return self.add_operation_and_number('<=', number)

    def __gt__(self, number):
        self.compile_code_line()
        return self.add_operation_and_number('>', number)

    def __ge__(self, number):
        self.compile_code_line()
        return self.add_operation_and_number('>=', number)

    def __eq__(self, other):
        if self._simulation_context is not None:
            simlog.error('Equality assignment not allowed for event condition in MobsPy.\n'
                         'Please if necessary use ( >= ) & ( =< )')
        else:
            return id(self) == id(other)

    def __neg__(self, other):
        return id(self) != id(other)

    def __hash__(self):
        return hash(id(self))


class ReactingSpeciesComparator(SpeciesComparator):
    """
        This class implements the comparisons necessary for events and conditional durations for Reacting Species
        ex (A.a1 > 5)

        :param _simulation_context: (Simulation) current simulation under context
    """

    def __init__(self):
        super(ReactingSpeciesComparator, self).__init__()
        self._simulation_context = None

    def check_context(self):
        """
            Checks if a reacting species is under context and add it to the attribute self._simulation_context
        """
        for react_dict in self.list_of_reactants:
            if react_dict['object']._simulation_context is not None:
                self._simulation_context = react_dict['object']._simulation_context
                break
        else:
            self._simulation_context = None

    def add_operation_and_number(self, symbol, number):
        """
            Adds the symbol and number to create a MetaSpeciesLogicResolver object

            :param symbol: Comparative symbols '<=', '<', '>=', or '>'
            :param number: (int) if compared to a value, (Species or ReactingSpecies) if compared to the objects

            :returns: MetaSpeciesLogicResolver object containing the comparison
        """
        number = self.reformat_number_and_species(number)
        if type(number) == list:
            operation = self.logical_add_reacting_species(self) + [symbol] + number
        else:
            operation = self.logical_add_reacting_species(self) + [symbol] + [number]
        logic_re_object = MetaSpeciesLogicResolver(operation, self._simulation_context)
        return logic_re_object


class MetaSpeciesLogicResolver:
    """
        This object stores the logical and comparison operations that become the event triggers for conditions
        or the conditional duration for simulations

        :param operation: (list) the logical operation that will be transformed in a string for copasi. Format ex
        [{object: ..., characteristics:...}, '<=', 10]
        :param simulation_context: (Simulation) simulation under context
    """

    def __init__(self, operation, simulation_context=None):
        self.operation = operation
        self.simulation_context = simulation_context

    def __and__(self, other):
        return self._join(other, '&&')

    def __or__(self, other):
        return self._join(other, '||')

    def _join(self, other, symbol):
        if not isinstance(other, MetaSpeciesLogicResolver):
            print('ERROR')
            exit()

        new_operation = ['('] + ['('] + self.operation + [')'] + [symbol] + ['('] + other.operation + [')'] + [')']
        self.operation = new_operation
        return self

    def __lt__(self, number):
        self.add_double_symbol('<', number)
        return self

    def __le__(self, number):
        self.add_double_symbol('<=', number)
        return self

    def __gt__(self, number):
        self.add_double_symbol('>', number)
        return self

    def __ge__(self, number):
        self.add_double_symbol('>=', number)
        return self

    def add_double_symbol(self, symbol, number):
        if type(number) != int or type(number) != float or not isinstance(number, Quantity):
            self.operation = [number, symbol] + self.operation

    @classmethod
    def find_all_species_strings(cls, species, characteristics, species_for_sbml):
        reference_set = set(characteristics)
        reference_set.add(str(species))
        return [x for x in species_for_sbml.keys() if reference_set.issubset(set(x.split('_dot_')))]

    def generate_string(self, characteristics_to_object, to_sort=False):
        copasi_str = ''
        for i, e in enumerate(self.operation):
            if type(e) == int or type(e) == float or type(e) == str:
                copasi_str = copasi_str + str(e) + ' '
            else:
                copasi_str = copasi_str + '('
                ite = ssg.construct_all_combinations(e['object'], e['characteristics'],
                                                     characteristics_to_object, '_dot_')
                if to_sort:
                    ite = sorted(ite)

                for j, species_str in enumerate(ite):
                    if j == 0:
                        copasi_str = copasi_str + f'{species_str}'
                    else:
                        copasi_str = copasi_str + f' + {species_str}'
                copasi_str = copasi_str + ')' + ' '

        return copasi_str



import sys, os
import matplotlib.pyplot as plt
from mobspy import *
import time


"""
    This is a geometry-based model
    Here the Meta-Species Mesh is used to represent the positions in a grid
    And their transitions using a CRN
    Here we have a bacteria and a phage that move through a Mesh and once they encounter the bacteria becomes infected
    Although the movement is not realistic, it shows MobsPy geometry capabilities 
    In future work we hope to implement different type movements or grids
"""

start_time = time.time()

Mesh = BaseSpecies(1)
n = 10
for i in range(n):
    for j in range(n):
        coordinate = 'p_' + str(i) + '_' + str(j)

        if i + 1 < n:
            Mesh.c(coordinate) >> Mesh.c(f'p_{i+1}_{j}')[0.1]
        if i - 1 > -1:
            Mesh.c(coordinate) >> Mesh.c(f'p_{i-1}_{j}')[0.1]
        if j - 1 > -1:
            Mesh.c(coordinate) >> Mesh.c(f'p_{i}_{j-1}')[0.1]
        if j + 1 < n:
            Mesh.c(coordinate) >> Mesh.c(f'p_{i}_{j+1}')[0.1]

Bacteria, Phage = New(Mesh, 2)
Bacteria.not_infected + Phage >> Bacteria.infected [lambda r1, r2: 1000000 if Mesh(r1) == Mesh(r2) else 0]
Bacteria.p_0_0(1)
Phage.c(f'p_{n-1}_{n-1}')(1)

MySim = Simulation(Bacteria | Phage)
MySim.simulation_method = 'stochastic'
MySim.repetitions = 1
MySim.duration = 100
MySim.output_event = True
MySim.plot_data = False
MySim.save_data = False

MySim.run()


def grab_position(species_position, list_x, list_y):
    _, x, y = Mesh(species_position).split('_')
    list_x.append(int(x))
    list_y.append(int(y))


data = MySim.results['data']

Bacteria_x = []
Bacteria_y = []
Phage_x = []
Phage_y = []
Meeting_x = []
Meeting_y = []
break_flag = False
for t in range(len(data['Time'])):
    position_flag_bacteria = False
    position_flag_phage = False

    for key in data:
        if key == 'Time' or key == 'Bacteria' or key == 'Phage':
            continue
        species_string = deepcopy(key)

        if data[key]['runs'][0][t] == 1:

            if species_string.split('.')[0] == 'Bacteria':
                if Bacteria(key) == 'infected':
                    grab_position(key, Meeting_x, Meeting_y)
                    break_flag = True
                    break

                if not position_flag_bacteria:
                    grab_position(key, Bacteria_x, Bacteria_y)
                    position_flag_bacteria = True
                else:
                    raise TypeError('Something went wrong')

            if species_string.split('.')[0] == 'Phage':
                if not position_flag_phage:
                    grab_position(key, Phage_x, Phage_y)
                    position_flag_phage = True
                else:
                    raise TypeError('Something went wrong')

    if break_flag:
        break

plt.plot(Bacteria_x, Bacteria_y, c='b')
plt.plot(Phage_x, Phage_y, c='r')
plt.plot(Meeting_x, Meeting_y, c='y')
plt.show()



"""
    The meta_class.py model is responsible for defining the meta-species, meta-reactions and other "meta-classes"
    It is also responsible for compiling the models into data structures that will be converted into an SBML file later
    Bear in mind that they are not actually Python meta-classes. The first design utilized this feature but know there
    are just regular classes
"""
import mobspy.modules.meta_class_utils as mcu
from mobspy.modules.function_rate_code import *
import mobspy.modules.reaction_construction_nb as rc
import mobspy.modules.function_rate_code as frc
from pint import Quantity
from mobspy.modules.order_operators import *
import mobspy.modules.event_functions as eh
from mobspy.modules.logic_operator_objects import *
import mobspy.modules.species_string_generator as ssg
from copy import deepcopy
import re


# Easter Egg: I finished the first version on a sunday at the BnF in Paris
# If anyone is reading this, I highly recommend you study there, it is quite a nice place
class Compiler:
    """
        The compiler is responsible for constructing the model from the meta-species and meta-reactions given
        and checking if it is a valid MobsPy model
    """

    # Basic storage of defined variables for Compiler
    entity_counter = 0
    ref_characteristics_to_object = {}
    last_rate = None

    @classmethod
    def override_get_item(cls, object_to_return, item):
        """
            Due to priority in Python the item is stored before the reaction
            So it is stored in the Compiler level and passed to the reaction object in the end
            Important: the rate is used before the compilation level, it's used when the reaction has been completely
            defined

            :param object_to_return: (Species or Reacting) returns the object that the __getitem__ was performed on
            :param item: (int, float, callable, Quantity) stored reaction rate
        """
        cls.last_rate = item
        return object_to_return

    @classmethod
    def add_phantom_reactions(cls, reactions_for_sbml, species_to_add_phantom_reaction):
        """
            This method is here because basico cannot compile species that have been assigned to events but are not
            present in reactions.
            Interestingly enough this error does not happen in python 3.10.
            However, to be compatible with earlier versions we add an extremely slow phantom reaction for each one
            of those species with a coefficient of 1e-100. The idea is that the reaction is so slow that it should
            not make any sizable difference in any model. It adds them directly to the reactions for sbml dictionary

            :param reactions_for_sbml: reactions dictionary in python sbml format
            :param species_to_add_phantom_reaction: species that will be assigned a phantom reaction
        """
        for i, spe in enumerate(species_to_add_phantom_reaction):
            reactions_for_sbml['phantom_reaction_' + str(i)] = {'re': [(1, spe)], 'pr': [],
                                                                'kin': str(spe) + ' * 1e-100'}

    @classmethod
    def compile(cls, meta_species_to_simulate, reactions_set, species_counts, orthogonal_vector_structure,
                volume=1, type_of_model='deterministic', verbose=True, event_dictionary=None,
                continuous_sim=False, ending_condition=None):
        """
            Here we construct the species for Copasi using their objects

            It handles the construction of species_for_sbml, reactions_for_sbml, parameters_for_sbml,
            mappings_for_sbml, model_str which will be used to write the sbml_string for basiCO
            It also performs checks to see if the model is valid

            :param meta_species_to_simulate: (List_Species or Species) Species to generate the model
            :param reactions_set: (set) Set of meta-reactions collected when the simulation objected was constructed
            :param species_counts: (dict) All counts assigned to species in the model
             before the simulation object was constructed
            :param orthogonal_vector_structure: (dict) ref_characteristics_to_objects in other modules. Dictionary
            with the characteristics as keys and the objects as values. Characteritics pointing to their coordinate
            in the vector space
            :param volume: (int, flot) Simulation volume
            :param type_of_model: (str) deterministic or stochastic
            :param verbose: (bool) print or not the model results by generating a model_str
            meta-reaction ( or Default we have the round-robin) and how MobsPy will handle meta-species in
            the products that are not referenced in the reactants - see order_operators.py
            :param event_dictionary: (dict) dictionary with all the events to be added to the model
            :param continuous_sim: (bool) simulation with conditional duration or not
            :param ending_condition: (MetaSpeciesLogicResolver) trigger for the ending contion

            :returns: species_for_sbml (dict) = dictionary where the species strings (in MobsPy format) are the keys and
            the values are their counts inside the model. reactions_for_sbml (dict) = Reaction in dictionary format
            with keys 're', 'pr' and 'rate'. Values are strings parameters_for_sbml (dict) = value with parameter
            name and tuple with quantity and unit. For standard models it only holds the volume mappings_for_sbml
            (dict) = Dictionary that maps a meta-species to the set of species strings that belong to it. model_str
            (str) = str of the variables defined above in a user-readable format. events_for_sbml (dict) = dictionary
            containing the event trigger and count assignments. assigned_species (list) = list of meta-species which
            had counts assigned to them during this model execution (used when composing simulations)
        """
        # Check to see if all species are named
        names_used = set()
        # Removing repeated elements to ensure only one meta-species in the simulation
        meta_species_to_simulate = meta_species_to_simulate.remove_repeated_elements()
        black_listed_names = {'Time', 'Rev', 'All'}
        for i, species in enumerate(meta_species_to_simulate):
            if species.get_name() in black_listed_names:
                simlog.error(f'The name {species.get_name()} is not allowed for meta-species please change it')
            if '$' in species.get_name():
                simlog.error('An error has occurred and one of the species was not named')
            if species.get_name() in names_used:
                simlog.error(f'Names must be unique for all species\n' +
                             f'The repeated name is {species.get_name()} in position {i}\n' +
                             f'Another possibility could be a repeated meta-species in the model')
            names_used.add(species.get_name())

        # Order the species references for later usage
        for species in meta_species_to_simulate:
            species.order_references()

        # Start by creating the Mappings for the SBML
        # Convert to user friendly format as well
        mappings_for_sbml = {}
        for spe_object in meta_species_to_simulate:
            mappings_for_sbml[spe_object.get_name()] = []

        # List of Species objects
        species_for_sbml = {}
        mappings_for_sbml = {}
        for spe_object in meta_species_to_simulate:
            species_string_list = ssg.construct_all_combinations(spe_object, 'std$',
                                                                 orthogonal_vector_structure)
            for x in species_string_list:
                x[0] = x[0].get_name()
            mappings_for_sbml[spe_object.get_name()] = ['.'.join(x) for x in species_string_list]
            for species_string in species_string_list:
                species_for_sbml['_dot_'.join(species_string)] = 0

        # Dimension check. Here we check based on the units the area dimension
        dimension = None
        for reaction in reactions_set:
            reaction_order = len(reaction.reactants)
            if isinstance(reaction.rate, Quantity):
                if uh.extract_length_dimension(str(reaction.rate.dimensionality), dimension, reaction_order):
                    dimension = uh.extract_length_dimension(str(reaction.rate.dimensionality), dimension,
                                                            reaction_order)

        for count in species_counts:
            if isinstance(count['quantity'], Quantity):
                if uh.extract_length_dimension(str(count['quantity'].dimensionality), dimension):
                    dimension = uh.extract_length_dimension(str(count['quantity'].dimensionality), dimension)

        # Check volume:
        if isinstance(volume, Quantity):
            uh.extract_length_dimension(str(volume.dimensionality), dimension)
        volume = uh.convert_volume(volume, dimension)

        # Add the flag species used for verifying if the simulation is over
        if continuous_sim:
            species_for_sbml[EndFlagSpecies.get_name()] = 0

        # Assignments with all do not take priority
        # This allows specific assignments to override All assignments
        assigned_species = []
        for count in species_counts:

            if 'all$' not in count['characteristics']:
                continue

            temp_set = set(count['characteristics'])
            temp_set.remove('all$')
            species_strings = ssg.construct_all_combinations(count['object'], temp_set,
                                                             orthogonal_vector_structure,
                                                             symbol='_dot_')

            temp_count = uh.convert_counts(count['quantity'], volume, dimension)
            for spe_str in species_strings:
                if type(temp_count) == float and not type_of_model == 'deterministic':
                    simlog.warning('The stochastic simulation rounds floats to integers')
                    species_for_sbml[spe_str] = int(temp_count)
                    assigned_species.append(spe_str)
                else:
                    species_for_sbml[spe_str] = temp_count
                    assigned_species.append(spe_str)

        # Set initial counts for SBML
        # Create the list HERE
        for count in species_counts:

            if 'all$' in count['characteristics']:
                continue

            species_string = ssg.construct_species_char_list(count['object'], count['characteristics'],
                                                             orthogonal_vector_structure,
                                                             symbol='_dot_')

            temp_count = uh.convert_counts(count['quantity'], volume, dimension)
            if type(temp_count) == float and not type_of_model == 'deterministic':
                simlog.warning('The stochastic simulation rounds floats to integers')
                species_for_sbml[species_string] = int(temp_count)
                assigned_species.append(species_string)
            else:
                species_for_sbml[species_string] = temp_count
                assigned_species.append(species_string)

        # BaseSpecies reactions for SBML with theirs respective parameters and rates
        # What do I have so far
        # Species_String_Dict and a set of reaction objects in Reactions_Set
        reactions_for_sbml, parameters_for_sbml = rc.create_all_reactions(reactions_set,
                                                                          meta_species_to_simulate,
                                                                          orthogonal_vector_structure,
                                                                          type_of_model, dimension)
        parameters_for_sbml['volume'] = (volume, f'dimensionless')

        # O(n^2) reaction check for doubles
        for i, r1 in enumerate(reactions_for_sbml):
            for j, r2 in enumerate(reactions_for_sbml):
                if i == j:
                    continue

                if reactions_for_sbml[r1]['re'] == reactions_for_sbml[r2]['re'] \
                        and reactions_for_sbml[r1]['pr'] == reactions_for_sbml[r2]['pr'] \
                        and reactions_for_sbml[r1]['kin'] == reactions_for_sbml[r2]['kin']:
                    simlog.warning('The following reaction: \n' +
                                   f'{reactions_for_sbml[r1]} \n' +
                                   'Is doubled. Was that intentional? \n')

        # Event implementation here
        # Basico does not resolve species that have no reactions but were added to events
        # So we add "phantom" reactions to fix this
        events_for_sbml, species_in_events = eh.format_event_dictionary_for_sbml(species_for_sbml, event_dictionary,
                                                                                 orthogonal_vector_structure,
                                                                                 volume, dimension,
                                                                                 meta_species_to_simulate)
        species_in_reactions = set()
        for key, reaction in reactions_for_sbml.items():
            for reactant in reaction['re']:
                species_in_reactions.add(reactant[1])
            for product in reaction['pr']:
                species_in_reactions.add(product[1])
        cls.add_phantom_reactions(reactions_for_sbml, species_in_events.difference(species_in_reactions))

        if continuous_sim:
            end_event = {'trigger': ending_condition.generate_string(orthogonal_vector_structure),
                         'delay': '0',
                         'assignments': [('End_Flag_MetaSpecies', '1')]}

            reactions_for_sbml['phantom_reaction_end'] = {'re': [(10, 'End_Flag_MetaSpecies')], 'pr': [],
                                                          'kin': 'End_Flag_MetaSpecies * 1e-100'}
            events_for_sbml['end_event'] = end_event

        model_str = ''
        if verbose:
            model_str = '\n'
            model_str += 'Species' + '\n'
            species_alpha = list(sorted(species_for_sbml.keys()))
            for spe in species_alpha:
                model_str += spe.replace('_dot_', '.') + ',' + str(species_for_sbml[spe]) + '\n'

            model_str += '\n'
            model_str += 'Mappings' + '\n'
            mappings_alpha = list(sorted(mappings_for_sbml.keys()))
            for map in mappings_alpha:
                model_str += map + ' :' + '\n'
                for element in sorted(mappings_for_sbml[map]):
                    model_str += element + '\n'

            model_str += '\n'
            model_str += 'Parameters' + '\n'
            parameters_alpha = list(sorted(parameters_for_sbml.keys()))
            for par in parameters_alpha:
                model_str += par + ',' + str(parameters_for_sbml[par][0]) + '\n'

            model_str += '\n'
            model_str += 'Reactions' + '\n'
            remove_phantom_reactions = deepcopy(reactions_for_sbml)
            to_remove = []
            for reaction in remove_phantom_reactions:
                if 'phantom' in reaction:
                    to_remove.append(reaction)
            for r in to_remove:
                remove_phantom_reactions.pop(r, None)
            reaction_alpha = [str(x[1]).replace('_dot_', '.') for x in
                              list(sorted(remove_phantom_reactions.items(), key=lambda x: str(x[1])))]

            for i, reac in enumerate(reaction_alpha):
                model_str += 'reaction_' + str(i) + ',' + reac + '\n'

            if events_for_sbml != {}:
                model_str += '\n'
                model_str += 'Events' + '\n'
                list_to_sort = [str(events_for_sbml[key]) for key in events_for_sbml]
                list_to_sort = sorted(list_to_sort)
                for i in range(len(list_to_sort)):
                    model_str += ('event_' + str(i) + ',' + list_to_sort[i] + '\n').replace('_dot_', '.')

        return species_for_sbml, reactions_for_sbml, parameters_for_sbml, mappings_for_sbml, model_str, \
               events_for_sbml, assigned_species


class Reactions:
    """
        This is the Reaction class. It contains the reactants, products, rate and order
        Reactions are created with the >> operator. In each reaction is stored in all involved objects

        :param reactants: (list of Species) list of meta-species used as reactants in the meta-reaction
        :param products: (list of Species) list of meta-species used as products in the meta-reaction
        :param order: (Order Operator) reaction order operator - default in a Round-Robin base
        :param rate: (int, float, callable, Quantity) reaction rate
    """

    @staticmethod
    def __create_reactants_string(list_of_reactants):
        """
            Just a simple way to simlog.debug reactions for debbuging
            Not relevant for simulation
            Important: here reactants are used interchangeably with products, this works for a list_of_products too

            :param list_of_reactants: (list of Species) list of products or reactants in the reaction
        """
        reaction_string = ''
        for i, r in enumerate(list_of_reactants):
            if r['stoichiometry'] > 1:
                reaction_string += str(r['stoichiometry']) + '*' + str(r['object'])
            else:
                reaction_string += str(r['object'])

            reaction_string += '.' + '.'.join(r['characteristics'])

            if i != len(list_of_reactants) - 1:
                reaction_string += ' '
                reaction_string += '+'
                reaction_string += ' '

        return reaction_string

    def __str__(self):
        """
            Prints the meta-reaction in the format A + B -> C + D
        """
        return self.__create_reactants_string(self.reactants) + \
               ' -> ' + self.__create_reactants_string(self.products)

    def __getitem__(self, item):
        """
            Override of __getitem__ for dealing with reaction rates

            :param item: (int, float, callable, Quantity) = reaction rate
        """
        return Compiler.override_get_item(self, item)

    def __init__(self, reactants, products):
        """
            Constructor of the reaction object. For the object construction only the reactants and products are
            necessary - the order and the rate are assigned later by the compiler

            :param reactants: (list of Species) list of meta-species used as reactants in the meta-reaction
            :param products: (list of Species) list of meta-species used as products in the meta-reaction

            :raise simlog.error: if a reaction is defined under a simulation context
            :raise simlog.error: if was called from Zero to Zero
        """
        try:
            to_test_context_object = reactants[0]['object']
        except IndexError:
            try:
                to_test_context_object = products[0]['object']
            except IndexError:
                simlog.error('No Meta-Species detected in the reaction', stack_index=3)

        if to_test_context_object._simulation_context is not None:
            simlog.error('Reactions cannot be defined under event context. Only species counts', stack_index=3)

        self.reactants = reactants
        self.products = products

        # Assign default order
        self.order = None

        self.rate = Compiler.last_rate
        if Compiler.last_rate is not None:
            Compiler.last_rate = None

        # Here we extract all involved objects to pact them in a set
        # This is done to find the reactions associated with the species when the Compiler is started
        for reactant in reactants:
            reactant['object'].add_reaction(self)
        for product in products:
            product['object'].add_reaction(self)

    def set_rate(self, rate):
        """
            Sets the stored reaction rate. See Compiler.override_get_item(self, item)

            :param rate: (int, float, callable, Quantity) = reaction rate
        """
        self.rate = rate


class Reacting_Species(ReactingSpeciesComparator):
    """
        This is a intermediary object created when a species is used in a reaction. It is created when a species is
        part of a reaction, so it's constructor is called by the Species class in several of it's methods. It
        transforms a Species object into a list packable object to pass to create the reaction
        object. This list format allows for easy concatenation of reactants by summing

        The >> operator calls the Reaction constructor to finally define a reaction object

        :param  list_of_reactants: (list of dict) used interchangeably for products and reactants.
        A list of dictionaries containing the following {'object': meta-species object reference,
        'characteristics': the characteristics queried, 'stoichiometry': stoichiometry value,
        'label': label if used (None)}
    """

    def __str__(self):
        """
            String representation of the list of reactants
        """
        species_object = self.list_of_reactants[0]['object']
        characteristics = self.list_of_reactants[0]['characteristics']
        if len(self.list_of_reactants) == 1:
            if species_object._simulation_context is None:
                to_return = str(species_object)
                for cha in self.list_of_reactants[0]['characteristics']:
                    to_return += '.' + cha
                return to_return
            else:
                return Species.str_under_context(species_object, characteristics)
        else:
            if species_object._simulation_context is not None:
                simlog.error('Please separate the species when using string based assignments under event context'
                             'Ex: str(A) + str(B)', stack_index=2)
            return str(self.list_of_reactants)

    # Labels and value function implementation
    def c(self, item):
        """
            c query implementation, queries by the value inside a variable instead of the name
            it just calls __getattr__ with the value inside the variable

            :param item: value to query over
        """
        item = str(item)
        Species.check_if_valid_characteristic(item)
        return self.__getattr__(item)

    def label(self, label):
        """
            Label function implementation. This function assigns labels to meta-species to be matched by the compiler
            later

            :param label: (int, float, str) value for the label for matching
        """
        if len(self.list_of_reactants) == 1:
            self.list_of_reactants[0]['label'] = label
        else:
            simlog.error('Labels cannot be assigned to multiple reacting species at the same time.', stack_index=2)
        return self

    def __getitem__(self, item):
        """
            Override of __getitem__ for dealing with reaction rates

            :param item: (int, float, callable, Quantity) = reaction rate
        """
        return Compiler.override_get_item(self, item)

    def __init__(self, object_reference, characteristics, stoichiometry=1, label=None):
        """
            Reacting_Species constructor. It receives the meta-species object reference, the characteristics that
            have been used as a query in the reaction, the stoichiometry of the meta-species in the reaction, and
            finally a label if used.

            :param object_reference: (Species) meta-species object reference or (int, float, Quantity) for event
            assignments with meta-species values
            :param characteristics: (str) characteristics used to query over the meta-species inside this reaction
            :param stoichiometry: (int) stoichiometry value of the meta-species in the reaction
            :param label: (int, float, str) value for the label for matching used in this reaction
        """
        super(Reacting_Species, self).__init__()
        if object_reference.get_name() == 'S0' and characteristics == set():
            self.list_of_reactants = []
        else:
            self.list_of_reactants = [{'object': object_reference, 'characteristics': characteristics,
                                       'stoichiometry': stoichiometry, 'label': label}]

    def __rmul__(self, stoichiometry):
        """
            Multiplication by the stoichiometry for reactions

            :params stoichiometry: (int) stoichiometry value of the meta-species in the reaction
        """
        if type(stoichiometry) == int:
            self.list_of_reactants[0]['stoichiometry'] = stoichiometry
        else:
            simlog.error(f'Stoichiometry can only be an int - Received {stoichiometry}', stack_index=2)
        return self

    def __add__(self, other):
        """
            Addition of meta-species to construct the reaction. Here we can add species ( by transforming into reacting
            species ) or other reacting species. With this we concatenate different reacting species in the
            list_of_reactants to eventually transform into a reaction with the >> operator

            :param  other: (Species or Reacting Species) other object being added
        """
        if isinstance(other, Species):
            other = Reacting_Species(other, set())
        try:
            self.list_of_reactants += other.list_of_reactants
        except AttributeError:
            simlog.error(f'Addition between meta-species and types {type(other)} is not supported', stack_index=2)
        return self

    def __radd__(self, other):
        return Reacting_Species.__add__(self, other)

    def __rshift__(self, other):
        """
            The >> operator for defining reactions. It passes two instances of reacting species to construct the
            list_of_reactants and list_of_products in the reaction object

            :param other: (Species or Reacting Species) product side of the reaction being added
        """
        code_line = inspect.stack()[1].code_context[0][:-1]
        line_number = inspect.stack()[1].lineno
        Species._compile_defined_reaction(code_line, line_number)

        if isinstance(other, Species):
            p = Reacting_Species(other, set())
        else:
            p = other

        reaction = Reactions(self.list_of_reactants, p.list_of_reactants)
        return reaction

    def __call__(self, quantity):
        """
            The call operator here is used to add counts to species non-default state. This stores the characteristics
            that have been called using the dot operator to assign the count after the call operation

            :param quantity: (int, float, Quantity) count to be assigned to the species
        """
        species_object = self.list_of_reactants[0]['object']
        simulation_under_context = self.list_of_reactants[0]['object']._simulation_context

        if type(quantity) == int or type(quantity) == float or isinstance(quantity, Quantity):
            if len(self.list_of_reactants) != 1:
                simlog.error('Assignment used incorrectly. Only one species at a time', stack_index=2)
            species_object = species_object
            characteristics = self.list_of_reactants[0]['characteristics']
            quantity_dict = species_object.add_quantities(characteristics, quantity)
        elif type(quantity) == Reacting_Species and simulation_under_context is not None:
            simlog.error(f'Assignments of counts using meta-species are only allowed under events in '
                         f'simulation context', stack_index=2)
        else:
            simlog.error(f'Reactant_Species count assignment does not support the type {type(quantity)}',
                         stack_index=2)

        if simulation_under_context is not None:
            try:
                if type(quantity) == str:
                    quantity_dict = self.add_quantities('std$', quantity)
                simulation_under_context.current_event_count_data.append({'species': species_object,
                                                                          'characteristics': quantity_dict[
                                                                              'characteristics'],
                                                                          'quantity': quantity_dict[
                                                                              'quantity']})
            except Exception as e:
                simlog.error(str(e) + '\n Only species count assignments are allowed in a model context')
        else:
            return self

    def __getattr__(self, characteristic):
        """
            This is the implementation of the .dot operation. Adds characteristics to the species and/or perform
            queries

            :param characteristic: (str) characteristic for the query
        """
        Species.check_if_valid_characteristic(characteristic)

        for reactant in self.list_of_reactants:

            species_object = reactant['object']
            characteristics_from_references = mcu.unite_characteristics(species_object.get_references())

            if characteristic not in characteristics_from_references and '$' not in characteristic:
                species_object.add_characteristic(characteristic)

            reactant['characteristics'].add(characteristic)

        return self

    @classmethod
    def is_species(cls):
        return False

    @classmethod
    def is_spe_or_reac(cls):
        return True


_methods_Reacting_Species = set(dir(Reacting_Species))


class List_Species:
    """
        This class just stores species after the | operation. It creates a list of species than can be loop through or
        given to the simulator

        :param list_of_species: (Species) Meta-species list to store the meta-species
    """

    def __init__(self, iterable):
        """
            Constructor not usually used - but a list_of_species is given one can construct the List_Species from it
            It is advised to use the | operator

            :param list_of_species: (Species) Meta-species list to store the meta-species
        """
        self._list_species = []
        for item in iterable:
            if isinstance(item, Species):
                self._list_species.append(item)
            else:
                simlog.error('Only Species can used to construct List_Species', stack_index=2)

    def append(self, species):
        """
            Add species to the list_of_species

            :param species: (Species) meta-species to be added to the List_Species
        """
        if not isinstance(species, Species):
            simlog.error('Only Species can be appended', stack_index=2)
        else:
            self._list_species.append(species)

    def __str__(self):
        """
            String representation of the List_Species just the list of meta-species
        """
        to_return = []
        for spe in self:
            to_return.append(spe.get_name())
        return str(to_return)

    def __or__(self, other):
        """
            Implementation of the | operator. Adds species to the Parallel species

            Parameters:
                other (Species) = meta-species to be added to the List_Species
        """
        if isinstance(other, Species):
            self._list_species.append(other)
        elif isinstance(other, List_Species):
            self._list_species = self._list_species + other._list_species
        else:
            simlog.error('Operator must only be used in Species on List_Species', stack_index=2)

        return self

    def __iter__(self):
        for spe in self._list_species:
            yield spe

    def __len__(self):
        return len(self._list_species)

    def is_in(self, item):
        if item in self._list_species:
            return True
        else:
            return False

    def __setitem__(self, index, item):
        self._list_species[index] = item

    def __getitem__(self, item):
        return self._list_species[item]

    def remove_repeated_elements(self):
        """
            This function creates a List_Species with no repetitions. The order of elements is lost

            :return: List_Species with no repeated elements
        """
        set_species = set(self._list_species)
        return List_Species(set_species)

    def insert(self, index, item):
        self._list_species.insert(index, item)

    def extend(self, other):
        if isinstance(other, List_Species):
            self._list_species = self._list_species + other._list_species

    def remove(self, value):
        """
            Removes all instances of the meta-species in the object

            :return: The number of instances removed
        """
        indexes = []
        for i, e in enumerate(self._list_species):
            if e == value or str(e) == str(value):
                indexes.append(-len(self._list_species) + i)

        for i in indexes:
            del self._list_species[i]
        return len(indexes)


def ListSpecies(number_of_elements, inherits_from=None):
    code = inspect.stack()[1].code_context[0][:-1]
    before_eq, after_eq = code.split('=')
    before_eq = before_eq.replace(" ", '')
    if ',' in before_eq:
        simlog.error(f"At: {code} \n"
                     f"No comas are allowed in the right-side of the equality during the creation of a ListSpecies")
    name = before_eq.split('=')[0]

    temp_list = []
    for i in range(number_of_elements):
        temp_name = name + '_' + str(i + 1)

        if inherits_from is None:
            temp_list.append(BaseSpecies([temp_name]))
        else:
            temp_list.append(New(inherits_from, [temp_name]))

    return List_Species(temp_list)


class Species(SpeciesComparator):
    """
        Fundamental class - The meta-species object
        Contains the characteristics, the species name, the reactions it is involved in
        and finally the other species it references
        Objects store all the basic information necessary to create an SBML file and construct a model
        So we construct the species through reactions and __getattr__ to form a model

        :param _name: (str) name of the species - named firstly as N$Counter as a placeholder. During compilation it is
        named by MobsPy to be equal to the variable name. It can also be named by the user using one the methods
        :param _characteristics: (str) set of characteristics DIRECTLY added to a species, does not contain inherited
        characteristics
        :param _references: (set) set of meta-species a meta-species has inherited from (used in it's construction).
        Every species contains itself in this set
        :param first_characteristic: (str) first characteristic added to the species
        :param _reactions: (set) Every species stores all reactions is involved in. The compiler performs the union of
        the set with all species and species in the _references set to get the reactions back
        :param _species_counts: (list) counts listed for the species. Stores dictionaries with the characteristics as
        keys and the counts of values
    """

    @classmethod
    def check_if_valid_characteristic(cls, char):
        """
            Checks if the characteristics is valid. If there is any conflict between the methods of the class or
            attributes of the class the user will be asked to rename the characteristic

            :param: char - name of the characteristic to be added:
            :raises: simlog.error if the characteristic name is not allowed
            :return: True if characteristic is allowed false if not
        """
        black_list = {'list_of_reactants', 'first_characteristic'}
        if char[0] == '_':
            simlog.error(f'Characteristic name {char} is not allowed. Please pick another name', stack_index=3)

        if char in _methods_Reacting_Species or char in _methods_Species or char in black_list:
            simlog.error(f'Characteristic name {char} is not allowed. Please pick another name', stack_index=3)
            return False
        else:
            return True

    @classmethod
    def str_under_context(cls, species_object, characteristics):
        """
            Returns the str representation of a species under context or not

            :param species_object: Meta-species object to construct the string
            :param characteristics: Characteristics to filter the strings
            :return: String in format (A_dot_a1 + A_dot_a2 + ....) with the parenthesis
        """
        ref_char_to_spe_obj = species_object._simulation_context.orthogonal_vector_structure
        all_strings = sorted(ssg.construct_all_combinations(species_object, characteristics,
                                                            ref_char_to_spe_obj, '_dot_'))
        to_str = all_strings[0]
        for i, e in enumerate(all_strings):
            if i == 0:
                continue
            else:
                to_str = to_str + ' + ' + e
        to_str = '(' + to_str + ')'
        return to_str

    def __str__(self):
        """
            String representation, just returns the species name
        """
        if self._simulation_context is None:
            return self._name
        else:
            return Species.str_under_context(self, 'std$')

    # Def c to get the value
    def c(self, item):
        """
            c query implementation, queries by the value inside a variable instead of the name
            it just calls __getattr__ with the value inside the variable

            :param item: value to query over
        """
        item = str(item)
        Species.check_if_valid_characteristic(item)
        return self.__getattr__(item)

    # Def labels
    def label(self, label):
        """
            Label function implementation. This function matches equal meta-species if they have equal labels
            Here it returns a Reacting_Species as labels need to be used in reactions

            :param label: (int, float, str) value for the label for matching
            :return: Reacting_Species object created with the label in the constructor
        """
        return Reacting_Species(self, set(), label=label)

    # Get data from species for debugging Compiler
    def show_reactions(self):
        """
            Prints the reactions inside the object
        """
        simlog.debug(str(self) + '_dot_')
        for reference in self._references:
            for reaction in reference.get_reactions():
                simlog.debug(reaction)

    def show_characteristics(self):
        """
            Prints the directly characteristics inside the object
        """
        simlog.debug(str(self) + ' has the following characteristics referenced:')
        for i, reference in enumerate(self.get_references()):
            if reference.get_characteristics():
                simlog.debug(str(reference) + ': ', end='')
                reference.simlog.debug_characteristics()

    def show_references(self):
        """
            Prints the objects this object has inherited from (including self)
        """
        simlog.debug(str(self) + '_dot_')
        simlog.debug('{', end='')
        for i, reference in enumerate(self.get_references()):
            if reference.get_characteristics():
                simlog.debug(' ' + str(reference) + ' ', end='')
        simlog.debug('}')

    def show_quantities(self):
        """
            Shows the species counts stored in this object
        """
        simlog.debug(self._species_counts)

    # Creation of List_Species For Simulation ##################
    def __or__(self, other):
        """
            Creates an instance of List_Species using the | operator

            :param other: (Species or List_Species) Species or List_Species to combine into
        """
        if isinstance(other, List_Species):
            other.append(self)
            return other
        elif isinstance(other, Species):
            return List_Species([self, other])
        else:
            simlog.error('Only Species and List_Species can be concatenated', stack_index=2)

    # Both are defined bellow to be consistent with List_Species behavior
    def __iter__(self):
        """
            iter defined to be consistent with List_Species behavior
            :return: Itself
        """
        yield self

    def remove_repeated_elements(self):
        """
            Defined to be consistent with the model behavior from List_Species
            :return: Itself
        """
        return self

    # Creation of reactions using entities ########################
    def __getitem__(self, item):
        """
            Override of __getitem__ for dealing with reaction rates

            :param item: (int, float, callable, Quantity) reaction rate
        """
        return Compiler.override_get_item(self, item)

    def __rmul__(self, stoichiometry):
        """
            Multiplication by the stoichiometry

            :param stoichiometry: (int) Stoichiometry of the meta-species in the meta-reaction
            :return: r (Reacting_Species) Reacting_Species with the stoichiometry added to it
        """
        if type(stoichiometry) == int:
            r = Reacting_Species(self, set(), stoichiometry)
        else:
            simlog.error(f'Stoichiometry can only be an int - Received {stoichiometry}', stack_index=2)
        return r

    def __add__(self, other):
        """
            Implementation of addition. In case a Species object is being used in a reaction, so it can then generate
            a Reacting_Species object

            :param other: (Species or Reacting Species object) Other objected added to construct a reaction
            :return: r1 + r2 (Reacting Species) Reacting Species objected created by the sum of the two
        """
        r1 = Reacting_Species(self, set())
        if isinstance(other, Reacting_Species):
            r2 = other
        else:
            r2 = Reacting_Species(other, set())
        return r1 + r2

    def __radd__(self, other):
        """
            Just making addition symmetric check __add__
        """
        return Species.__add__(self, other)

    @classmethod
    def _compile_defined_reaction(cls, code_line, line_number):
        n_all = code_line.count('All')
        if 'Rev' in code_line:
            n_all = n_all + 1

        n_key = code_line.count(']')
        if n_key != n_all + 1:
            simlog.error(f'At: {code_line} \n' + f'Line number: {line_number} \n'
                        + f'The reaction did not compile. Perhaps there is a missing rate?')

        if code_line[-1] != ']':
            simlog.error(f'At: {code_line} \n' + f'Line number: {line_number} \n'
                         + f'There must be a rate in the end of the reaction')

    def __rshift__(self, other):
        """
            Reaction definition (>> operator) in case one uses only a single species. Allows it to transform
            into a reacting species before adding it into the reaction

            :param other: (Species or Reacting_Species) reaction products
            :return: the reaction
        """
        myself = Reacting_Species(self, set())
        code_line = inspect.stack()[1].code_context[0][:-1]
        line_number = inspect.stack()[1].lineno
        Species._compile_defined_reaction(code_line, line_number)

        if isinstance(other, Species):
            p = Reacting_Species(other, set())
        elif other == 0:
            exit()
        else:
            p = other

        reaction = Reactions(myself.list_of_reactants, p.list_of_reactants)
        return reaction

    # Adding characteristics to an Entity or Property ####################
    def __getattr__(self, characteristic):
        """
            This adds characteristics using the following manner Species.characteristic
            If it is the first time is called we add it to the set of characteristics
            Second time, it just creates a Reacting_Species for reaction construction

            :param characteristic: (str) characteristic to be added or to be use as a query in the reaction
            :return: Reacting_Species with the characteristic added for querying
        """
        Species.check_if_valid_characteristic(characteristic)

        characteristics_from_references = mcu.unite_characteristics(self.get_references())
        characteristics = {characteristic}

        if characteristic not in characteristics_from_references and '$' not in characteristic:
            if len(self.get_characteristics()) == 0:
                self.first_characteristic = characteristic

            self.add_characteristic(characteristic)

        return Reacting_Species(self, characteristics)

    # Adding counts to species
    def __call__(self, quantity):
        """
            The __call__ operator handles two things for Species objects
            First, it adds characteristics to the default state of the meta-species if the quantity is a real
            Secondly, it returns the name of the characteristic from a species string that belongs to this meta-species

            :param: quantity (int, float, Quantity) = for count assignment
            (Specific_Species_Operator object) = for characteristic extraction

            :return self: to allow for assigning counts mid-reaction
        """
        if type(quantity) == int or type(quantity) == float or isinstance(quantity, Quantity):
            quantity_dict = self.add_quantities('std$', quantity)
        elif isinstance(quantity, frc.Specific_Species_Operator):
            for cha in str(quantity).split('_dot_')[1:]:
                if cha in self._characteristics:
                    return cha
            simlog.error(f'{quantity} contains no characteristics from {self._name}', stack_index=2)
        elif type(quantity) == Reacting_Species:
            simlog.error(f'Assignments of counts using meta-species are only allowed under events in '
                         f'simulation context', stack_index=2)

        if self._simulation_context is not None:

            if type(quantity) == str:
                quantity_dict = self.add_quantities('std$', quantity)
            try:
                self._simulation_context.current_event_count_data.append({'species': self,
                                                                          'characteristics': quantity_dict[
                                                                              'characteristics'],
                                                                          'quantity': quantity_dict['quantity']})
            except Exception as e:
                simlog.error(str(e) + '\n Only species count assignments are allowed in a model context')
        else:
            return self

    def add_quantities(self, characteristics, quantity):
        """
            This function was implemented because Python can't accept sets as dictionary keys
            Set the quantity of an specific string of species

            :param characteristics: (str) characteristics of the species to be set
            :param quantity: (int, float, Quantity) counts of that specific species
        """
        if self._simulation_context is None:
            already_in = False
            for e in self._species_counts:
                if characteristics == e['characteristics']:
                    e['quantity'] = quantity
                    already_in = True
            if not already_in:
                self._species_counts.append({'characteristics': characteristics, 'quantity': quantity})
        else:
            return {'characteristics': characteristics, 'quantity': quantity}

    def reset_quantities(self):
        """
            Just reset the counts inside a species - potentially useful in jupyter environment
        """
        self._species_counts = []

    def get_quantities(self):
        """
            Returns:
                The list of species_counts
        """
        return self._species_counts

    # Creation of other objects through multiplication ##################
    def __mul__(self, other):
        """
            Multiplications are used to construct more complex species
            We do not concatenate characteristics. This keeps the structure intact for reaction construction
            Instead we combine the sets of references, objects can access the characteristics from other through the
            reference set

            :param other: (Species) Species for the multiplication and creation of a higher order species
            :return: new_entity (Species) Higher order species resulted from the multiplication
        """
        code_line = inspect.stack()[1].code_context[0][:-1]

        if not isinstance(other, Species):
            simlog.error(f'At {code_line}: \n' + 'Meta-Species can only be multiplied by other meta-species \n'
                         + f'It was multiplied by the type {type(other)}')

        name = code_line.replace(" ", "").split("=")[0]

        Compiler.entity_counter += 1
        new_entity = Species(name)
        new_entity.set_references(mcu.combine_references(self, other))
        new_entity.add_reference(new_entity)

        mcu.check_orthogonality_between_references(new_entity.get_references())

        return new_entity

    def __init__(self, name):
        """
            Object constructor - We recommend using BaseSpecies instead

            :param name: (str) Name of the species (can be placeholder if named with N$)
        """
        super(Species, self).__init__()
        self._name = name
        self._characteristics = set()
        self._references = {self}
        self._ordered_references = []
        self._simulation_context = None
        self._reference_index_dictionary = {}

        # This is necessary for the empty objects generated when we perform multiplication with more than 2 Properties
        self.first_characteristic = None

        # Each object stores the reactions it is involved in
        self._reactions = set()

        # This will store the quantities relating to the species counts
        self._species_counts = []

    def name(self, name):
        """
            Function to name a species. Allows the user to use a name different from the variable's name

            :param name: (str) name of the species
        """
        self._name = name

    def get_name(self):
        return self._name

    def get_characteristics(self):
        return self._characteristics

    def add_characteristic(self, characteristic):
        self._characteristics.add(characteristic)

    def remove_characteristic(self, characteristic):
        self._characteristics.remove(characteristic)

    def print_characteristics(self):
        simlog.debug(self._characteristics)

    def get_references(self):
        return self._references

    def get_all_characteristics(self):
        """
            Different that get_characteristics. This function the characteristics directly added to it and the
            characteristics of the references. While get_characteristics only returns the characteristics directly
            added to that meta-species.
        """
        all_char = set()
        for reference in self._references:
            all_char = all_char.union(reference.get_characteristics())
        return all_char

    def add_reference(self, reference):
        self._references.add(reference)

    def set_references(self, reference_set):
        self._references = reference_set

    def reset_references(self):
        self._references = {self}

    def get_reactions(self):
        return self._reactions

    def set_reactions(self, reactions):
        self._reactions = reactions

    def reset_reactions(self):
        self._reactions = set()

    def add_reaction(self, reaction):
        self._reactions.add(reaction)

    def reset_counts(self):
        self._species_counts = []

    def set_simulation_context(self, sim):
        if self._simulation_context is None:
            self._simulation_context = sim
        else:
            simlog.error('A different Simulation Object was assigned to a meta-species object under context \n'
                         'Please use only one Simulation Object per context assignment', stack_index=6)

    def reset_simulation_context(self):
        self._simulation_context = None

    def order_references(self):
        cleaned_references = [x for x in self.get_references() if x.get_characteristics() != set()]
        self._ordered_references = sorted(cleaned_references, key=lambda x: sorted(list(x.get_characteristics())))
        i = 1
        for reference in self._ordered_references:
            self._reference_index_dictionary[reference] = i
            i = i + 1

    def get_ordered_references(self):
        return self._ordered_references

    def get_index_from_reference_dict(self, reference):
        return self._reference_index_dictionary[reference]

    @classmethod
    def is_species(cls):
        return True

    @classmethod
    def is_spe_or_reac(cls):
        return True


_methods_Species = set(dir(Species))


def compile_species_number_line(code_line):
    """
        Compiles code line for BaseSpecies and New

        :param code_line: - Line of code where BaseSpecies or New has been called
        :return: n (int) = number of variables after a coma, names (list) = list of strings with the names of the
        variables
    """
    before_eq, after_eq = code_line.split('=')[0], code_line.split('=')[1]
    if after_eq.count('BaseSpecies') > 1:
        simlog.error(f'At {after_eq}: \n' + 'BaseSpecies can only be called once at a time')

    code_line = before_eq
    n = code_line.count(',') + 1
    code_line = code_line.replace(' ', '')
    names = code_line.split(',')

    return n, names


def _Create_Species(species, code_line, number_or_names=None):
    if number_or_names is not None:
        if type(number_or_names) == int:
            if number_or_names < 1:
                simlog.error('Please use strictly positive integers for number of properties', stack_index=3)
        elif type(number_or_names) == list:
            pass
        else:
            simlog.error('Only numbers or lists of strings accepted', stack_index=3)

    if number_or_names is None or type(number_or_names) == int:
        number_of_properties, compiled_names = compile_species_number_line(code_line)
        names = compiled_names
        if number_or_names is not None:
            if number_of_properties != number_or_names:
                simlog.error(f'The number of properties is not equal to the number of variables', stack_index=3)
    elif type(number_or_names) == list:
        number_of_properties = len(number_or_names)
        names = number_or_names

    to_return = []
    for i in range(number_of_properties):
        Compiler.entity_counter += 1
        if names is None:
            name = 'N$' + str(Compiler.entity_counter)
        else:
            name = names[i]
        if species is None:
            to_return.append(Species(name))
        else:
            temp = __S1 * species
            temp.name(name)
            to_return.append(temp)

    if len(to_return) == 1:
        return to_return[0]
    else:
        return tuple(to_return)


# Property Call to return several properties as called
def BaseSpecies(number_or_names=None):
    """
        Function that returns a number of base species according to the number_of_properties
        BaseSpecies are just Species with no inheritance attached to it

        :param number_or_names: (int) number of base species wanted or (list) list of names of wanted Species
        :return: Species objects according to the number of species requested
    """
    code_line = inspect.stack()[1].code_context[0][:-1]
    return _Create_Species(None, code_line, number_or_names)


__S0, __S1, __SF = BaseSpecies(3)
__S0.name('S0')
__S1.name('S1')
__SF.name('End_Flag_MetaSpecies')
EndFlagSpecies = __SF
Zero = __S0


def New(species, number_or_names=None):
    """
        Function that returns a number of meta-species that inherit from the species supplied

        :param number_or_names: (int) number of meta-species wanted, (list) list of names wanted
        :return: Species objects according to the number of species requested
    """
    code_line = inspect.stack()[1].code_context[0][:-1]
    return _Create_Species(species, code_line, number_or_names)


if __name__ == '__main__':
    pass



"""
    This module implements the class that stores the results from a MobsPy simulation
"""
import pandas as pd
from mobspy.modules.meta_class import *
import inspect

class MobsPyTimeSeries:

    def __init__(self, data_dict):
        """Creates the MobsPy timeseries object

            :param data: (dict) resulting dictionary from simulation
            {'data': ...., 'params':....., 'models':.......}
        """
        self.ts_data = data_dict['data']
        self.ts_parameters = data_dict['params']
        self.ts_models = data_dict['models']

    def to_dict(self):
        """
            :return: data in dict format {'data': ...., 'params':....., 'models':.......}
        """
        return {'data': self.ts_data,
                'params': self.ts_parameters,
                'models': self.ts_models}

    def __len__(self):
        """
            Length of MobsPy Time Series is equal to the number of runs from simulation
        """
        return len(self.ts_data)

    def __str__(self):
        tr = ''
        for i, data in enumerate(self.ts_data):
            tr = tr + f'Time Series {i}: \n'
            tr = tr + str(pd.DataFrame.from_dict(data)) + '\n\n'
        return tr

    def add_ts_to_data(self, time_series):
        """
            Add a new time series to the TS data. Used for stochastic plotting
            the average and standard deviation

            :param time_series: (dict) dictionary with species strings as keys and run as value
        """
        if type(time_series) == dict:
            self.ts_data += [time_series]

    def __getitem__(self, item):
        """
            Implements run retrieval using a meta-species object. Returns one run if there is only one
            time-series and returns multiple runs if there are multiple time series
        """
        code_line = inspect.stack()[1].code_context[0][:-1]
        if '[\'runs\']' in code_line:
            simlog.error('As of version 2.0.1 MobsPy has changed the data output format. \n'
                         'Now data can be accessed through the following syntax: \n'
                         'S.results[Meta-Species Object] or S.results[Meta-Species string name] \n'
                         'Both can perform queries', stack_index=2)

        to_return = []
        if type(item) == int:
            return self.ts_data[item]
        elif type(item) == str:
            try:
                for ts in self.ts_data:
                    to_return.append(ts[item])
            except KeyError:
                for i in range(len(self)):
                    to_return.append(self._sum_reacting_species_data(item, i))
        elif isinstance(item, Species):
            for ts in self.ts_data:
                to_return.append(ts[item.get_name()])
        elif isinstance(item, Reacting_Species):
            for i in range(len(self)):
                to_return.append(self._sum_reacting_species_data(item, i))

        if len(to_return) == 1:
            return to_return[0]
        else:
            return to_return

    def _sum_reacting_species_data(self, item, ts_index):
        """
            Maps meta-species according to characteristics ex: A.a1 = A.a1.b1 + A.a1.b2 + A.a1.b3

            :param item: Meta-species object or string to be retrieved
            :ts_index: Index of the time series to perform the sum
        """
        def _sum_element_by_element(l1, l2):
            rt = []
            for e1, e2 in zip(l1, l2):
                rt.append(e1 + e2)
            return rt

        time_series = self.ts_data[ts_index]
        to_return = [0 for _ in range(len(time_series['Time']))]

        found_flag = False
        if isinstance(item, Reacting_Species):
            for reactant in item.list_of_reactants:
                for key in time_series:
                    reactant_name = reactant['object'].get_name()
                    reactant_full_name = reactant['characteristics'].union([reactant_name])
                    if reactant_full_name.issubset(set(key.split('.'))):
                        to_return = _sum_element_by_element(to_return, time_series[key])
                        found_flag = True
        elif type(item) == str:
            for key in time_series:
                if set(item.split('.')).issubset(set(key.split('.'))):
                    to_return = _sum_element_by_element(to_return, time_series[key])
                    found_flag = True

        if not found_flag:
            simlog.error(f'{item} was not found in data')

        return to_return

    def __iter__(self):
        if len(self.ts_data) > 1:
            for ts in self.ts_data:
                yield ts
        else:
            for key in self.ts_data[0]:
                yield key

    def get_max_time_for_species(self, species):
        """
            Returns the maximum time in all the time-series stored for a given species
        """
        if type(species) != str:
            species = species.get_name()
        max_length = 0
        max_ts = None
        for ts in self.ts_data:
            if species in ts.keys():
                if len(ts) > max_length:
                    max_length = len(ts)
                    max_ts = ts
        return max_ts['Time']



import numpy as np
import mobspy.simulation_logging.log_scripts as simlog


def time_series_average(species_string, mobspy_ts):
    """
        Badly named function - Average between all RUNS inside a single time-series

        :param species_string: (str) species string to perform the average upon
        :param mobspy_ts: (MobsPy TimeSeries) MobsPy time series object
        :return: average_series (list) a list with the average values from all runs
    """

    list_series = []
    for series in mobspy_ts:
        if species_string in series.keys():
            list_series.append(series)

    for s1 in list_series:
        for s2 in list_series:
            if s1 == s2:
                continue
            # if len(s1['Time']) != len(s2['Time']):
            #    simlog.warning('The time length vectors sizes are different \n')

            for t1, t2 in zip(s1['Time'], s2['Time']):
                if t1 != t2:
                    simlog.error('Time vectors are different')

    average_series = []
    for j in range(len(mobspy_ts.get_max_time_for_species(species_string))):
        add = 0
        size = 0
        for series in list_series:
            try:
                add = add + series[species_string][j]
                size += 1
            except IndexError:
                pass
            except KeyError:
                pass

        average_series.append(add/size)

    return average_series


def standard_deviation(species_string, mobspy_ts, average_series = None):
    """
        Standard deviation between all RUNS inside a single time-series

        :param species_string: (str) Species string
        :param mobspy_ts: (MobsPy TimeSeries) MobsPy time series object
        :param average_series: (list) if the average series is given it is not recalculated
        :return: deviation_series (list) a list with the standard deviation values from all runs
   """
    if average_series is None:
        average_series = time_series_average(species_string, mobspy_ts)
    deviation_series = []

    for j in range(len(mobspy_ts.get_max_time_for_species(species_string))):

        add = 0
        size = 0
        for series in mobspy_ts:
            try:
                add = add + (average_series[j] - series[species_string][j])**2
                size += 1
            except IndexError:
                pass
            except KeyError:
                pass

        deviation_series.append(np.sqrt(add/size))

    return deviation_series


def average_plus_standard_deviation(species_string, mobspy_ts, average_series=None, deviation_series=None):
    """
        Standard deviation between all RUNS inside a single time-series

        :param species_string: (str) Species string
        :param mobspy_ts: (MobsPy TimeSeries) MobsPy time series object
        :param average_series: (list) if the average series is given it is not recalculated
        :param deviation_series: (list) if the deviation is given it is not recalculated

        :return: series_average (list) = average value of the run,  plus (list) = average + deviation,
        minus (list) = average - deviation,
   """
    if average_series is None:
        series_average = time_series_average(species_string, mobspy_ts)
    if deviation_series is None:
        series_deviation = standard_deviation(species_string, mobspy_ts)

    plus = []
    minus = []
    for average, deviation in zip(series_average, series_deviation):
        plus.append(average + deviation)
        minus.append(average - deviation)

    return series_average, plus, minus



import mobspy.plot_scripts.statistics_calculations as sc
from copy import deepcopy
import mobspy.plot_scripts.hierarchical_plot as hp
import json
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.plot_scripts.process_plot_data as ppd


def read_plot_json(plot_json_filename):
    """
        This function converts a plot_json file into a dictionary

        :param plot_json_filename: (str) JSON file name

        :return: json_data (dict) converted JSON as dictionary
    """
    with open(plot_json_filename, 'r') as file:
        try:
            json_data = json.load(file)
        except Exception:
            simlog.error(f'Error while decoding json file "{plot_json_filename}".')

    return json_data


def set_plot_units(new_plot_params):
    """
        Sets the plot labels to the unit names by adding to the xlabel and ylabel

        :param new_plot_params: (dict) plot parameters after some changes
    """
    new_plot_params['xlabel'] = 'Time'
    if new_plot_params['unit_x'] is not None:
        new_plot_params['xlabel'] += f'({new_plot_params["unit_x"]})'
    else:
        new_plot_params['xlabel'] += '(s)'

    if new_plot_params['output_concentration']:
        new_plot_params['ylabel'] = 'Conc.'
    else:
        new_plot_params['ylabel'] = 'Counts'

    if new_plot_params['unit_y'] is not None:
        new_plot_params['ylabel'] += f'({new_plot_params["unit_y"]})'
    else:
        if new_plot_params['output_concentration']:
            new_plot_params['ylabel'] += '(mol/L)'


def stochastic_plot(species, data, plot_params):
    """
        Design default stochastic plot using MobsPy plotting hierarchy. It them passes the parameters for plotting
        in the hierarchical plot module

        :param species: (list of str) list of species names in MobsPy str format (queries are performed with the
        query plot data function)
        :param data: (dict) data in MobsPy format Simulation.results['data']
        :param plot_params: (dict) dictionary with the plot parameters supplied by the user before the modifications
        from this function
    """

    # Data Handling - Copy data object to not interfere with simulation data
    species, data = ppd.query_plot_data(species, data)
    ppd.check_plot_parameters(species, plot_params)

    data_to_plot = data

    new_plot_params = deepcopy(plot_params)
    set_plot_units(new_plot_params)

    # Plot config
    new_plot_params['frameon'] = False
    new_plot_params['figures'] = []
    new_plot_params['pad'] = 1.5
    color_cycler = hp.Color_cycle()
    for spe in species:
        # We define new 'mappings' with the resulting runs for the statics for the plot structure
        try:
            plots_for_spe_i = []
            plots_for_spe_i_sta = []

            processed_runs = sc.average_plus_standard_deviation(spe, data_to_plot)

            key_average = spe + '$' + 'average'
            key_dev = spe + '$' + 'deviation'

            average_and_deviation_ts = {'Time': data_to_plot.get_max_time_for_species(spe),
                                        key_average: processed_runs[0],
                                        key_dev: [processed_runs[1], processed_runs[2]]}
            data_to_plot.add_ts_to_data(average_and_deviation_ts)

            # We define the standard plot for the average and deviation
            color = color_cycler(1)
            plots_for_spe_i.append({'species_to_plot': [spe],
                                    spe: {'color': color, 'label': spe}})

            plots_for_spe_i_sta.append({'species_to_plot': [key_average],
                                        key_average: {'color': color, 'linestyle': '-', 'label': 'mean'}})
            plots_for_spe_i_sta.append({'species_to_plot': [key_dev], 'fill_between': True,
                                        key_dev: {'color': (0.8, 0.8, 0.8), 'linestyle': ':', 'label': 'std. dev'}})
        except ValueError:
            simlog.error(f'{spe} species not found in data')
        new_plot_params['figures'].append({'ylabel': spe + ' ' + new_plot_params['ylabel'], 'plots': plots_for_spe_i})
        new_plot_params['figures'].append(
            {'ylabel': spe + ' ' + new_plot_params['ylabel'], 'plots': plots_for_spe_i_sta})

    hp.plot_data(data_to_plot, new_plot_params)


def deterministic_plot(species, data, plot_params):
    """
        Design default deterministic plot using MobsPy plotting hierarchy. It them passes the parameters for plotting
        in the hierarchical plot module

        :param species: (list of str) list of species names in MobsPy str format (queries are performed
        with the query plot data function)
        :param data: (dict) data in MobsPy format Simulation.results['data']
        :param plot_params: (dict) dictionary with the plot parameters supplied by the user before the modifications
        from this function
    """
    # Data Handling
    species, data = ppd.query_plot_data(species, data)
    ppd.check_plot_parameters(species, plot_params)

    new_plot_params = deepcopy(plot_params)
    set_plot_units(new_plot_params)

    #  Plot Config
    new_plot_params['frameon'] = False
    new_plot_params['species_to_plot'] = species
    color_cycler = hp.Color_cycle()
    for spe in species:
        new_plot_params[spe] = {'label': spe, 'color': color_cycler(1)}
    hp.plot_data(data, new_plot_params)


def raw_plot(data, parameters_or_file):
    """
        Plots data from a json or parameter dictionary configured according to the hierarchical plot structure
        Does not accept parameters from a Simulation object, it must be given the parameters in it's entirety

        :param data: (dict) data in MobsPy format Simulation.results['data']
        :param parameters_or_file: (dict, str) Dictionary originated from a JSON or JSON file name
    """
    if type(parameters_or_file) == str and parameters_or_file[-5:] == '.json':
        plot_params = read_plot_json(parameters_or_file)
    elif type(parameters_or_file) == dict:
        plot_params = parameters_or_file
    else:
        simlog.error('Raw plot only takes json files or parameters for configuration')

    species = list(data.ts_data[0].keys())
    ppd.check_plot_parameters(species, plot_params)

    # Here we add some base data just in case the user has not supplied it
    hp.plot_data(data, plot_params)


if __name__ == '__main__':
    raw_plot({}, 'default_parameters.json')



"""
    This module is responsible for converting a model_str into a SBML format
"""
import libsbml as sbml


def check(value, message="error"):
    """If 'value' is None, prints an error message constructed using
    'message' and then exits with status code 1.  If 'value' is an integer,
    it assumes it is a libSBML return status code.  If the code value is
    LIBSBML_OPERATION_SUCCESS, returns without further action; if it is not,
    prints an error message constructed using 'message' along with text from
    libSBML explaining the meaning of the code, and exits with status code 1.
    """
    if value is None:
        raise RuntimeError(f"LibSBML returned a null value trying to {message}.")

    elif type(value) is int:
        if value == sbml.LIBSBML_OPERATION_SUCCESS:
            return
        else:
            err_msg = (
                "Error encountered trying to "
                + message
                + "."
                + "LibSBML returned error code "
                + str(value)
                + ': "'
                + sbml.OperationReturnValue_toString(value).strip()
                + '"'
            )
            raise RuntimeError(err_msg)
    else:
        return


def create_model(species={}, parameters={}, reactions={}, events={}):
    """
    Returns an SBML Level 3 model.
    Example:
    species = { 'E': 1,
                'EM': 0,
                'EM2': 0,
                'F': 100,
                },
    parameters = {'k': (1e-06,'per_min'),
                 }
    reactions = { 'Production_E':
                        { 're': [(1,'E'),(1,'F')],
                          'pr': [(2,'E')],
                          'kin' : 'k * E * F'
                        },
                },
    events = {'e':
              { 'trigger': 'true',
                'delay': '10',
                'assignments': [('M','1'),],
              },
    }
    """

    # Create an empty SBMLDocument object.  It's a good idea to check for
    # possible errors.  Even when the parameter values are hardwired like
    # this, it is still possible for a failure to occur (e.g., if the
    # operating system runs out of memory).

    try:
        document = sbml.SBMLDocument(3, 1)
    except ValueError:
        raise RuntimeError("Could not create SBMLDocumention object")

    # Create the basic Model object inside the SBMLDocument object.

    model = document.createModel()
    check(model, "create model")
    check(model.setTimeUnits("second"), "set model-wide time units")
    check(model.setExtentUnits("item"), "set model units of extent")
    check(
        model.setSubstanceUnits("item"), "set model substance units"
    )  # mole, item, gram, kilogram, dimensionless

    # Create a unit definition we will need later.

    per_second = model.createUnitDefinition()
    check(per_second, "create unit definition")
    check(per_second.setId("per_min"), "set unit definition id")
    unit = per_second.createUnit()
    check(unit, "create unit")
    check(unit.setKind(sbml.UNIT_KIND_SECOND), "set unit kind")
    check(unit.setExponent(-1), "set unit exponent")
    check(unit.setScale(0), "set unit scale")
    check(
        unit.setMultiplier(1), "set unit multiplier"
    )

    # Create a compartment inside this model

    c1 = model.createCompartment()
    check(c1, "create compartment")
    check(c1.setId("c1"), "set compartment id")
    check(c1.setConstant(True), 'set compartment "constant"')
    check(c1.setSize(1), 'set compartment "size"')
    check(c1.setSpatialDimensions(3), "set compartment dimensions")
    check(
        c1.setUnits("dimensionless"), "set compartment size units"
    )

    # Create species inside this model, set the required attributes
    # for each species in SBML Level 3 (which are the 'id', 'compartment',
    # 'constant', 'hasOnlySubstanceUnits', and 'boundaryCondition'
    # attributes), and initialize the amount of the species along with the
    # units of the amount.

    for s_str, s_val in species.items():
        s = model.createSpecies()
        check(s, "create species")
        check(s.setId(s_str), "set species id")
        check(s.setCompartment("c1"), "set species compartment")
        check(s.setConstant(False), 'set "constant" attribute')
        check(s.setInitialAmount(float(s_val)), "set initial amount")
        check(s.setSubstanceUnits("item"), "set substance units")
        check(s.setBoundaryCondition(False), 'set "boundaryCondition"')
        check(s.setHasOnlySubstanceUnits(False), 'set "hasOnlySubstanceUnits"')

    # Create a parameter object inside this model, set the required
    # attributes 'id' and 'constant' for a parameter in SBML Level 3, and
    # initialize the parameter with a value along with its units.

    for k_str in parameters:
        k = model.createParameter()
        check(k, "create parameter k")
        check(k.setId(k_str), "set parameter id")
        check(k.setConstant(True), 'set parameter "constant"')
        check(k.setValue(parameters[k_str][0]), "set parameter value")
        check(k.setUnits(parameters[k_str][1]), "set parameter units")

    # Create a reaction inside this model, set the reactants and products,
    # and set the reaction rate expression (the SBML "kinetic law").  We
    # set the minimum required attributes for all of these objects.  The
    # units of the reaction rate are determined from the 'timeUnits' and
    # 'extentUnits' attributes on the Model object.

    for r_str in reactions:
        r = model.createReaction()
        check(r, "create reaction")
        check(r.setId(r_str), "set reaction id")
        check(r.setReversible(False), "set reaction reversibility flag")
        check(r.setFast(False), 'set reaction "fast" attribute')

        reactants = reactions[r_str]["re"]
        for re_val, re_str in reactants:
            species_ref = r.createReactant()
            check(species_ref, "create reactant")
            check(species_ref.setSpecies(re_str), "assign reactant species")
            check(species_ref.setStoichiometry(re_val), "set set stoichiometry")
            check(species_ref.setConstant(True), 'set "constant" on species')

        products = reactions[r_str]["pr"]
        for pr_val, pr_str in products:
            species_ref = r.createProduct()
            check(species_ref, "create product")
            check(species_ref.setSpecies(pr_str), "assign product species")
            check(species_ref.setStoichiometry(pr_val), "set set stoichiometry")
            check(species_ref.setConstant(True), 'set "constant" on species')

        math_ast = sbml.parseL3Formula(reactions[r_str]["kin"])
        kinetic_law = r.createKineticLaw()
        check(math_ast, f"create AST for rate expression")
        check(kinetic_law, "create kinetic law")
        check(kinetic_law.setMath(math_ast), "set math on kinetic law")

    # create events
    for e_str in events:
        e = model.createEvent()
        check(e, "create event")
        check(e.setId(e_str), "set id")
        check(e.setUseValuesFromTriggerTime(False), "?")

        t = model.createTrigger()
        check(t, "create trigger")
        check(
            t.setMath(sbml.parseL3Formula(events[e_str]["trigger"])),
            "set trigger condition",
        )
        check(t.setPersistent(False), "default not persistent")
        check(t.setInitialValue(False), "default not initially true")
        check(e.getTrigger().getMath(), 'Problem when creating the trigger condition. The trigger will not work.')
        # print( '> ' + sbml.formulaToString(e.getTrigger().getMath()) )

        d = model.createDelay()
        check(d, "create delay")
        check(d.setMath(sbml.parseFormula(events[e_str]["delay"])), "set math")
        check(e.setDelay(d), "set delay")

        for ass in events[e_str]["assignments"]:
            ea = model.createEventAssignment()
            check(ea, "check event assignment")
            check(ea.setVariable(ass[0]), "set variable")
            check(ea.setMath(sbml.parseL3Formula(ass[1])), "set math")

    return document  # sbml.writeSBMLToString(document)



from mobspy import *

A, B, C = BaseSpecies(3)
A(200) + B(100) >> C[0.005]
MySim = Simulation(A | B | C)
MySim.save_data = False
MySim.plot_data = False

# The hierarchical plot structure was designed to facilitate the plotting of multiple
# figures and curves within a single plot window
#
# In this plotting structure there is a figures key that accepts a list.
# For each element in this list, the hierarchical plotting structure will create a new figure
# Each figure contains a plots list with the number of curves to be plotted inside the figure
#
# Now the hierarchical plot structure works in the following manner:
#   It follows the order:
#   => plot-wise, then figure-wise and finally global-wise
#   The configuration parameters can be placed inside the plots, figures or as a key to the dictionary itself
#   When plotting MobsPy will check to see if the parameter was defined for the curve
#   If it was defined for the curve the parameter is used for plotting
#   If it is not defined it checks the figure
#   Then it repeats the same process for figure-wise before moving to global-wise
#
# Also the plotting structure allows one to assign some parameters by species
# To do so just create a dictionary where the key is the species name as can be seen bellow
# Based on this, if you wish to use MobsPy plotting structure please avoid naming your species
# with plotting parameters.
# A viable strategy is to always capitalize species names as no parameter is capitalized
#
# Both plot_stochastic and plot_deterministic can be configured using this structure
# Plot_raw can only be configured through this structure and does not accept the dot-based configuration
example_parameters = {

    'A': {'color': 'red',  'label': 'A'},
    'B': {'color': 'blue',  'label': 'B'},
    'C': {'color': 'green', 'label': 'C'},
    "species_to_plot":["A"],

    'figures':[{
            "plots":[
                {"species_to_plot": ["B"]},
                {}
        ]},{
            "plots":[
                {"species_to_plot": ["C"]},
                {}
        ]}
    ]
}

MySim.run()
MySim.plot_raw(example_parameters)



import mobspy.modules.meta_class_utils as mcu
import mobspy.modules.unit_handler as uh
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.modules.species_string_generator as ssg
from pint import Quantity


def format_event_dictionary_for_sbml(species_for_sbml, event_list, characteristics_to_object,
                                     volume, dimension, meta_species_to_simulate):
    """
        Creates events_for_sbml dictionary for sbml file construction on the sbml_simulator/SBMLWriter.py

        :param species_for_sbml: (dict) {'species_string': count, ....}
        :param event_list: [{'species':'meta_species_object', 'characteristics':['list of characteristics'],
            'quantity': number or pint object}
        :param characteristics_to_object: {'characteristic':'meta_species_object', .....}
        :param volume: Simulation volume for unit conversion
        :param dimension: Dimension of the system 2D, 3D, 4D, e
        :return: event dictionary for the sbml file construction
        :rtype: events = {'e': { 'trigger': 'true', 'delay': '10', 'assignments': [('M','1'),]}}
    """
    reformed_event_list = []
    species_in_events = set()

    # Convert count from triggers
    for ev in event_list:
        if ev['trigger'] != 'true':
            for i, e in enumerate(ev['trigger'].operation):
                if isinstance(e, Quantity):
                    ev['trigger'].operation[i] = uh.convert_counts(e, volume, dimension)

    for ev in event_list:
        if not ev['event_counts']:
            continue
        event_dictionary = {}

        # All assignments never take priority over specific assignments
        for ec in ev['event_counts']:
            if 'all$' not in ec['characteristics']:
                continue

            temp_char = set(ec['characteristics'])
            temp_char.remove('all$')
            dummy = ssg.construct_all_combinations(ec['species'], temp_char,
                                                   characteristics_to_object, symbol='_dot_')
            for d in dummy:
                if type(ec['quantity']) != str:
                    event_dictionary[d] = uh.convert_counts(ec['quantity'], volume, dimension)
                else:
                    event_dictionary[d] = ec['quantity']

        for ec in ev['event_counts']:
            if 'all$' in ec['characteristics']:
                continue

            dummy = ssg.construct_species_char_list(ec['species'], ec['characteristics'],
                                                    characteristics_to_object, symbol='_dot_')
            if type(ec['quantity']) != str:
                event_dictionary[dummy] = uh.convert_counts(ec['quantity'], volume, dimension)
            else:
                event_dictionary[dummy] = ec['quantity']

        if type(ev['trigger']) == str:
            reformed_event_list.append({'event_time': ev['event_time'], 'event_counts': event_dictionary,
                                        'trigger': ev['trigger']})
        else:
            for e in ev['trigger'].operation:
                if type(e) == dict:
                    if e['object'] not in meta_species_to_simulate:
                        simlog.error(f'Meta species {e["object"]} was used in an event but is not in the model')
            reformed_event_list.append({'event_time': ev['event_time'],
                                        'event_counts': event_dictionary,
                                        'trigger': ev['trigger'].generate_string(characteristics_to_object,
                                                                                 to_sort=True)})

    events_for_sbml = {}
    for i, event in enumerate(reformed_event_list):
        assignments = []
        for key in event['event_counts']:
            if key in species_for_sbml:
                assignments.append((key, str(event['event_counts'][key])))
                species_in_events.add(key)
            else:
                simlog.error(f'Species {key} used in an event assignment but it is not in the model')

        assignments.sort()
        events_for_sbml['e' + str(i)] = {'trigger': event['trigger'],
                                         'delay': str(event['event_time']),
                                         'assignments': assignments}

    return events_for_sbml, species_in_events



from mobspy import *
"""
    This model was designed to explain how units can be added to a MobsPy project.
    The units are based on the pint module.
"""

# Defining base-species and give it some characteristics:
A = BaseSpecies(1)
A.a1, A.a2, A.a3, A.a4

# MobsPy adopts a Liter(decimeter)-Second-Count convention.
# We assume those units when a number is given without units.
# To add units use the u letter and the dot operator with the unit:
A.a1(1*u.mol)

# One can use concentrations to specify initial counts as well.
# They will be converted using the model volume:
A.a2(2*u.nanomolar)

# Rate constants can also use units.
# In this case the correctness of units for rate constants are checked:
# 1/time for one reactant, volume/time for 2 reactants, volume**2/time for 3 reactants ...
A.a1 + A.a2 >> A.a3 + A.a4 [1*u.meter**3/u.second]

# Functions that return rate constants can return units as well:
A >> A.a4 [lambda r1: 5/u.second if r1.a1 else 1/u.second]

# But functions that return strings have no support for units yet.
# They are implicitely supposed to return rates in terms of 1/second.

MySim = Simulation(A)

# To set the duration and the volume of the simulation with units we use:
MySim.volume = 1*u.nanolitre
MySim.duration = 3*u.hours

# The model compilation shows the values in MobsPy with standard units:
# Liter(decimeter)-Second-Count
print(MySim.compile())

"""
    A more advanced model for other unit types
"""

# In MobsPy one can also use 2-D concentrations and volumes.
# As an example one can refer to the For_The_Tress model in the application models.
# There, concentrations are expressed as Count/Area and rates are expressed in Area/Time.
# The volume would be an area in this case (Although the parameter is still named volume; it is a volume in a 2D-space).
# If an inconsistence in dimensions (e.g., mixture between 2-D and 3-D) is detected, the compiler will throw an error.

print()
print('Other Model')
print()

try:
    B = BaseSpecies(1)
    B(2*u.molar)
    MySim = Simulation(B)
    MySim.volume = 2*u.meter**2
    print(MySim.compile())
except SystemExit:
    pass






"""
    This module deals with order_operators
    Thus it deals with the Round-Robin order assignment and with species in the products that
    are not referenced in the reactants (Born Species)
    It has also the implementation of the Rev operator, since that is a also a reaction operator
"""
from copy import deepcopy
import mobspy.modules.meta_class_utils as mcu
import mobspy.modules.meta_class as mc
import mobspy.simulation_logging.log_scripts as simlog
import mobspy.modules.species_string_generator as ssg


class __Operator_Base:
    """
        This is the order operator base. It contains some functions that can be useful the reaction operators defined
        Reaction Operators are now responsible for order assignment and how to deal with species in the products that
        are not referenced in the reactants (Born Species)
        It also overrides the getitem method so we can use:
        Operator[Reaction] to assign an order to the reaction
    """

    # Assign order structure
    def __getitem__(self, item):
        try:
            if type(item) == str:
                return item + '.all$'

            if item.is_species():
                return item.c('all$')
            elif not item.is_species():
                for reactant in item.list_of_reactants:
                    reactant['characteristics'].add('all$')
                return item
        except AttributeError:
            simlog.error('All can only be used on species, reacting species and strings under set_count',
                         stack_index=2)

    # Transform product function
    @staticmethod
    def transform_species_string(species_string, characteristics_to_transform,
                                 ref_characteristics_to_object):
        """
            This function handles the characteristic change due to a chemical reaction. It receives a string in the
            MobsPy format and the characteristic one has written on the product side of the reaction
            MobsPy them converts the characteristics in the same position as that one (that have been directly
            added to the same meta-species) to the one in the characteristics_to_transform variable

            :param  species_string: (str) A string associated with a species in the MobsPy format
            :param  characteristics_to_transform: (str) Characteristics received by the species object in the product
            :param ref_characteristics_to_object: (dict) Dictionary with the characteristics as keys and meta-species
            objects as values
        """
        species_object = species_string[0]
        species_to_return = [species_object.get_name()] + deepcopy(species_string[1:]) \
            if len(species_string) > 1 else [species_object.get_name()]

        for characteristic in characteristics_to_transform:
            obj = ref_characteristics_to_object[characteristic]
            i = species_object.get_index_from_reference_dict(obj)
            species_to_return[i] = characteristic
        species_to_return = '_dot_'.join(species_to_return)

        return species_to_return

    @staticmethod
    def find_all_string_references_to_born_species(species_referenced_by, characteristics,
                                                   ref_characteristics_to_object):
        """
            Finds all the string species that reference the born-meta species, including inheritors
            Returns them all in a list for the product construction

            :param species_referenced_by: (meta-species list) List of species that inherit from the born-meta species
            in the reaction
            :param characteristics: (str) List of characteristics queried on the born species in the product
            :param meta_species_in_model: (list) list of meta-species in model
            as values
        """
        to_return = []
        for species in species_referenced_by:
            to_return += ssg.construct_all_combinations(species, characteristics,
                                                        ref_characteristics_to_object, symbol='_dot_')
        return to_return

    @staticmethod
    def find_all_default_references_to_born_species(species_referenced_by, characteristics,
                                                    ref_characteristics_to_object):
        """
            Finds only the DEFAULT string species that reference the born-meta species, including inheritors
            Returns them all in a list for the product construction.

            :param species_referenced_by: (meta-species list) List of species that inherit from the born-meta species
            in the reaction
            :param characteristics: (str) List of characteristics queried on the born species in the product
            :param meta_species_in_model: (list) list of meta-species in model
            :param ref_characteristics_to_object: (dict) Dictionary with the characteristics as keys and meta-species
            objects that they have been directly added to as values
        """
        to_return = []
        for species in species_referenced_by:
            to_return += [ssg.construct_species_char_list(species, characteristics,
                                                          ref_characteristics_to_object, symbol='_dot_')]

        return to_return

    def __call__(self, order_dictionary, product_species,
                 model, ref_characteristics_to_object, all_reactions=False):
        """
            This function is responsible for parrying the products with the reactants according to their meta-species
            It parries them using the order_dictionary and an index-system for the round-robin application
            If it cannot find a parrying it considers it a born species

            :param order_dictionary: (dict) Dictionary with the meta-species as keys and the list of 
            meta-species strings
            :param product_species: (dict) Product species dict with the meta-species object (key: species),
            label (key: label), and characteristics (key: characteristics)
            :param ref_characteristics_to_object: (dict) Dictionary with the characteristics as keys and meta-species
            objects that they have been directly added to as values
        """
        round_robin_index = {}
        for species, label in [(e['species'], e['label']) for e in product_species]:
            round_robin_index[(species, label)] = 0

        products = []
        for species, label, characteristics in [(e['species'], e['label'], e['characteristics']) for e in
                                                product_species]:

            if 'all$' in characteristics:
                species_is_referenced_by = []
                for spe_obe in model:
                    if species in spe_obe.get_references():
                        species_is_referenced_by.append(spe_obe)
                products.append(self.find_all_string_references_to_born_species(species_is_referenced_by,
                                                                                characteristics,
                                                                                ref_characteristics_to_object))
                continue

            # Simple round robin
            try:
                species_to_transform_string = order_dictionary[(species, label)][round_robin_index[(species, label)]]
                round_robin_index[(species, label)] = (round_robin_index[(species, label)] + 1) % len(
                    order_dictionary[(species, label)])

                # Return in list of lists format for combination later
                products.append([self.transform_species_string(species_to_transform_string, characteristics,
                                                               ref_characteristics_to_object)])

            # If the species is not on the reactants - order_dictionary
            except KeyError:
                species_is_referenced_by = []
                for spe_obe in model:
                    if species in spe_obe.get_references():
                        species_is_referenced_by.append(spe_obe)

                if len(species_is_referenced_by) == 0:
                    simlog.error(f'Species {species} was used in a reaction '
                                 f'but itself or any inheritors are not in the model. Please add at least one')

                if all_reactions:
                    # Find all the species that reference the one in the reaction
                    products.append(self.find_all_string_references_to_born_species(species_is_referenced_by,
                                                                                    characteristics,
                                                                                    ref_characteristics_to_object))
                else:
                    # Find only default state of the species
                    products.append(self.find_all_default_references_to_born_species(species_is_referenced_by,
                                                                                     characteristics,
                                                                                     ref_characteristics_to_object))

        return products


class __Round_Robin_Base(__Operator_Base):
    """
        Here we have the implementation of the round robin order it goes like this:

        2*Ecoli >> 4*Ecoli

        Since an Ecoli is a generic object that refers to all types of Ecoli
        This object refers to multiple reactions and to multiple strings
        Therefore we follow this rule to know which string goes where:
        1 Ecoli reactant >> 1 Ecoli product
        2 Ecoli reactant >> 2 Ecoli product
        1 Ecoli reactant >> 3 Ecoli product
        2 Ecoli reactant >> 4 Ecoli product
        This is a cycle (round robin) between the reactants to be assigned positions in the product
        The products will keep the characteristics of the reactants except if stated otherwise with .
        For completely new species (no reactant of the same species) we use ALL possible combinations
        For only the default option see the code bellow
    """

    def __call__(self, order_dictionary, product_species,
                 meta_species_in_model,
                 ref_characteristics_to_object, all_reactions=True):
        """
            This function is responsible for parrying the products with the reactants according to their meta-species
            It parries them using the order_dictionary and an index-system for the round-robin application
            If it cannot find a parrying it considers it a born species

            :param order_dictionary: (dict) Dictionary with the meta-species as keys and the list of 
            meta-species strings
            :param product_species: (dict) Product species dict with the meta-species object (key: species),
            label (key: label), and characteristics (key: characteristics)
            :param ref_characteristics_to_object: (dict)  Dictionary with the characteristics as keys and meta-species
            objects that they have been directly added to as values
        """
        return super().__call__(order_dictionary, product_species, meta_species_in_model,
                                ref_characteristics_to_object, all_reactions)


# Define class to override the operators
All = __Round_Robin_Base()


class __RR_Default_Base(__Operator_Base):
    """
        Only default options for born species (no reference in reactant)
        See __Round_Robin_Base for clarification
    """

    # Here is the default order requested by Thomas
    def __call__(self, order_dictionary, product_species,
                 meta_species_in_model,
                 ref_characteristics_to_object, all_reactions=False):
        """
            This function is responsible for parrying the products with the reactants according to their meta-species
            It parries them using the order_dictionary and an index-system for the round-robin application
            If it cannot find a parrying it considers it a born species

            :param order_dictionary: (dict) Dictionary with the meta-species as keys and the list of meta-species
            strings
            :param product_species: (dict) Product species dict with the meta-species object (key: species),
            label (key: label), and characteristics (key: characteristics)
            :param ref_characteristics_to_object: (dict) Dictionary with the characteristics as keys and meta-species
            objects that they have been directly added to as values
        """
        return super().__call__(order_dictionary, product_species, meta_species_in_model,
                                ref_characteristics_to_object, all_reactions)


Default = __RR_Default_Base()


class __Set_Reversible_Rate:
    """
        This class is responsible for dealing with reversible reactions
        Since reversible reactions have too rates
        The process is in __getitem__
    """

    def __getitem__(self, both_rates):
        """
            This function extracts both reaction rates from a tuple

            :param both_rates: (rate functions) rate function from the direct and reverse reaction
        """
        try:
            if len(both_rates) != 2:
                simlog.error('The reversible reaction must receive 2 rates', stack_index=2)
        except TypeError:
            simlog.error('The reversible reaction must receive 2 rates', stack_index=2)

        self.reaction_direct.rate = both_rates[0]
        self.reaction_reverse.rate = both_rates[1]

    def __init__(self):
        """
            This a dummy constructor, the class object Rev is the one with the __getitem__ overridden
        """
        self.reaction_direct = None
        self.reaction_reverse = None

    def set_reactions(self, reaction_direct, reaction_reverse):
        self.reaction_direct = reaction_direct
        self.reaction_reverse = reaction_reverse


# Reversible reaction operator
class __Reversible_Base:

    def __getitem__(self, reaction):
        """
            This __getitem__ uses the rate setter which is a instance of __Set_Reversible_Rate to set the reversible
            reaction rates. It also creates the reverse reaction with the constructor

            :param reaction: (Reaction object) object from the reaction class
        """
        reaction_direct = reaction
        reaction_reverse = mc.Reactions(reaction_direct.products, reaction_direct.reactants)
        self.rate_setter.set_reactions(reaction_direct, reaction_reverse)
        return self.rate_setter

    def __init__(self, rate_setter):
        """
            This a dummy constructor, the class object Rev is the one with the __getitem__ overridden
        """
        self.rate_setter = rate_setter


Rev = __Reversible_Base(__Set_Reversible_Rate())



"""
    This module is responsible for processing the parameters given to MobsPy before starting the simulation
"""
import json
import sys
from datetime import datetime
import os
from pathlib import Path
import numpy as np
import mobspy.simulation_logging.log_scripts as simlog
from pint import Quantity


def read_json(json_file_name):
    """
        Reads json file

        :param plot_json_filename: json file name

        :raise simlog.error: If the file was not able to be read

        :return: plot parameter dictionary
    """
    with open(json_file_name, 'r') as file:
        try:
            json_data = json.load(file)
        except json.decoder.JSONDecodeError as e:
            simlog.error('Error reading file')
            exit(1)

    return json_data


def __name_output_file(params):
    """
        Gives a name to the output file - just date time in case the user has not specified one

        :param params: (dict) Dictionary with simulation parameters
    """

    file_name = "r_"
    file_name += str(datetime.now()) + '.json'
    params["absolute_output_file"] = params["output_dir"] + file_name


def __check_stochastic_repetitions_seeds(params):
    """
        The list of seeds must be equal to the number of repetitions specified

        :param params (dict) = Dictionary with simulation parameters
    """
    if 'seeds' in params:
        try:
            if params['repetitions'] != len(params['seeds']):
                simlog.error('Seeds must be equal to the number of repetitions')
        except Exception:
            simlog.error('Parameter seeds must be a list')


def __convert_parameters_for_COPASI(params):
    """
        Converts parameters units to MobsPy standard units (basiCO needs seconds for simulation duration)

        :param params: (dict) = Dictionary with simulation parameters
    """
    for key, p in params.items():
        if isinstance(p, Quantity):
            if str(p.dimensionality) == '[time]':
                params[key] = p.to('second').magnitude
                continue


def parameter_process(params):
    __name_output_file(params)
    __check_stochastic_repetitions_seeds(params)
    __convert_parameters_for_COPASI(params)





from mobspy import *

'''
    Here we will how MobsPy handles results
    Both as a plotting aspect and result accessing aspect
'''

# Firstly define the model
A, B, C = BaseSpecies(3)
A(200) + B(100) >> C [0.001]
MySim = Simulation(A | B | C)
MySim.save_data = False

# Set to stochastic
MySim.simulation_method = 'stochastic'

# MobsPy usually has two default plots
# One for stochastic and one for deterministic
# The default plots will be called accordingly to the simulation_method
# after the simulation
# Unless plot_data is set to false
MySim.plot_data = False

# Run the sim
MySim.run()

# If the default plot was not used one can always call the plot_stochastic or plot_deterministic
# from the resulting simulation object
# These function take either species as arguments or nothing at all
# If no argument is passed all species will be plotted
# If some species are passed as argument, they will be the only ones to be plotted
# For the stochastic_plot the default plot consists of two figures for each species
# One for the runs and the other for the average value with standard deviation
MySim.plot_stochastic(A, C)

# If save_data is true MobsPy will generate a JSON file containing the data
# Otherwise it can be accessed from the simulation object
# The simulation results contain the data, parameters and mappings used in the model
# The data is separated by species name
# And stored with all the runs under the runs key
# For example to recover the first run for the A species:
print(MySim.results['data']['A']['runs'][0])

"""
    Plotting with queries
"""


# Plotting functions can also query
# For instance in the model bellow we are interested in the evolution of a species with a certain characteristic
# We want to observe all the values of C with the characteristic b1
# So we just pass C.b1 as an argument to the plotting function

A, B = BaseSpecies(2)
A.a1, A.a2
B.b1, B.b2
C = A*B
C.b1.a1(100), C.b1.a2(100)
C.b1 >> C.b2 [0.1]
MySim = Simulation(C)
MySim.save_data = False
MySim.plot_data = False
MySim.run()

MySim.plot_deterministic(C.b1)




from mobspy import *

"""
    In this example we will construct a model to explain the round-robin structure in detail.
    The round-robin is a cyclic assignment structure that helps MobsPy deal with non-linear reactions.
    We will use some examples to show its importance.
"""

# We define the species:
Age = BaseSpecies(1)
Animal = New(Age)

# We assign characteristics to a species:
Animal.a1, Animal.a2

# We define an age reaction:
Age.young >> Age.old [1]

# The Round-Robin mechanism is demonstrated with the reaction bellow:
#   In this reaction we want to define a competition reaction between a young and an old animal.
#   We would like the old animal to survive the competition.
#
#   However, how does MobsPy distinguish this case from the case where the
#   young animal survives and becomes old?
#
#   In MobsPy the order of the meta-species in the reaction is important.
#   With the usage of a rond-robin system the first reactant becomes the first-species in the product
#   and it continues in a cyclic fashion until all products have been mapped to a species.
Animal.old + Animal.young >> Animal.old [1]
# Here, the old animal is the one that survives.

# The .label() can be used to circumvent this default behavior, if required.
# In the next reaction, we have the young Animal killing the old and becoming old itself.
Animal.old + Animal.young.label(1) >> Animal.old.label(1) [1]
# Here, the young animal is the one that survives.

# While arguably, the fact that once the old animal survives and once the young one
# does not make a difference in our setting, it does if additional characteristics are
# added to the Animal, e.g., it having a color.

# The cyclic round robin style is especially useful in reproduction reaction.
# Since the assignment cycles in a round-robin fashion, the first product will be assigned the first reactant
# and the second product will cycle through the reactants and be assigned the first reactant as well.
# This results, in the example below, the two products to have the same characteristics as the single reactant.
Animal >> 2*Animal [1]

# We recommend commenting out one of the competition reactions and running the model
# to play with the round-robin system.
MySim = Simulation(Animal)
print(MySim.compile())









import sys, os
import matplotlib.pyplot as plt
from mobspy import *


"""
    This is a more classical oscillator
    Here the repression is implemented using a hill function as a string
    It's the classical circle of three compounds repressing each other
"""

Promoter, Chemical = BaseSpecies(2)
Promoter.inactive, Promoter.active
DNAPromoter = New(Promoter)
DNAPromoter.PLcl(100), DNAPromoter.PLacl(100), DNAPromoter.PTetR(100)
Chemical.TetR(1)

# Here we define the activation reactions for the list of chemicals and promoters
list_of_chemicals = ['TetR', 'Lcl', 'Lacl']
list_of_promoters = ['PLcl', 'PLacl', 'PTetR']
for che, pro in zip(list_of_chemicals, list_of_promoters):
    Rev[ DNAPromoter.inactive.c(pro) + Chemical.c(che) >> DNAPromoter.active.c(pro)][1, 1]
Chemical >> Zero [1]

# And here we define the repression cycle
# Also we use a string for a non-mass action kinetics rate, we want the hill function for repression here
repressed = ['TetR', 'Lcl', 'Lacl']
repressors = ['Lacl' ,'TetR', 'Lcl']
hill = lambda che: f'10/(1 + ({che})^3)'
for rpsor, rpsed in zip(repressed, repressors):
    Chemical.c(rpsor) >> Chemical.c(rpsed) + Chemical.c(rpsor) [hill]

MySim = Simulation(DNAPromoter | Chemical)
MySim.duration = 300
MySim.plot_data = False
MySim.save_data = False
MySim.run()
MySim.plot_deterministic(Chemical.TetR, Chemical.Lcl, Chemical.Lacl)






import sys, os
from mobspy import *

"""
    Here we have a model with a Phage, Donor and Receptor with resources being considered
    Both the Donor and Receptor can reproduce with resources
    The Donor can produce Phages with resources
    Only the Receptor can be infected by the phages
"""
# Variable declaration
Resource, Mortal, Infectible = BaseSpecies(3)

# Resource Definition
AA, Glu = New(Resource, 2)
AA(100), Glu(100)

# Donor and Phage Creation
Mortal >> Zero [0.1]
Donor, Phage = New(Mortal, 2)
Donor(100)

# Here both Resources can react with the Donor due to inheritance
dup_rate = lambda _,resource: 0.2 if resource.is_a(AA) else 0.1
Donor + Resource >> 2*Donor [dup_rate]
Donor + Resource >> Donor + Resource + Phage [0.1]
Infectible.low_inf >> Infectible.high_inf [0.1]
Receptor = Mortal*Infectible
inf_rate = lambda receptor: 0.2 if receptor.high_inf else 0.1
Receptor.not_infected + Phage >> Receptor.early_infection [inf_rate]
Receptor.early_infection >> Receptor.late_infection [0.1]
# We separate the reproduction reactions, because Receptor will give birth to a low_inf one
Receptor + Resource >> Receptor.low_inf + Receptor [dup_rate]

MySim = Simulation(Donor | Receptor | Phage | AA | Glu)
print(MySim.compile())



import sys, os
from mobspy import *

"""
    This is a basic MobsPy example
    Here we declare four species and the single reaction A + B >> 2*C + D
    It enunciates the basics of MobsPy, like declaring a species using the BaseSpecies constructor
    Species counts are assigned using the call method
    Rates are assigned using the getitem brackets []
"""
A, B, C, D = BaseSpecies(4)
A(200) + B(100) >> 2*C + D [0.1]
My_Sim = Simulation(A | B | C | D)
My_Sim.save_data = False
My_Sim.duration = 5
My_Sim.volume = 10
My_Sim.simulation_method = 'stochastic'
My_Sim.run()



# This script tests the model creation and compilation
# It also test the calculation capabilities
# It uses some simple model and assertions

# import pytest
from mobspy import *
import sys
import matplotlib.pyplot as plt

# TODO Plot has random order for species names


# Compare results with expected file
def compare_model(comp_results, file_name):
    with open(file_name, 'r') as file:
        for r, line in zip(comp_results.split('\n'), file.readlines()):
            line = line.replace('\n', '')
            if r != line:
                print(line)
                return False
    return True


# Model to test the basics
def test_model_1():
    A, B, C = BaseSpecies(3)
    A + B >> C[1]
    MySim = Simulation(A | B | C)
    MySim.level = -1
    results = MySim.compile()
    assert compare_model(results, 'test_tools/model_1.txt')


# Model to test basic inheritance
def test_model_2():
    Carnivore, Herbivore = BaseSpecies(2)
    Cat, Dog = New(Carnivore, 2)
    Carnivore + Herbivore(1 * u.mol) >> Carnivore[1]
    Cat(1 * u.mol), Dog(1 * u.mol)
    MySim = Simulation(Cat | Dog | Herbivore)
    MySim.level = -1
    MySim.volume = 1 * u.meter ** 2
    results = MySim.compile()
    assert compare_model(results, 'test_tools/model_2.txt')


# Model to test species multiplication
def test_model_3():
    MGMT, Blue_Oyster_Cult, The_Smiths = BaseSpecies(3)
    MGMT.eletric_fell, MGMT.little_dark_age, MGMT.kids
    Blue_Oyster_Cult.burning_for_you >> Blue_Oyster_Cult.reaper[1]
    The_Smiths.stop_me >> The_Smiths.charming_man[1]
    Music = MGMT * Blue_Oyster_Cult * The_Smiths
    MySim = Simulation(Music)
    MySim.level = -1
    results = MySim.compile()
    assert compare_model(results, 'test_tools/model_3.txt')


# Model to test inheritance queries
# All bacterias are infected by any virus here
def test_model_4():
    Bacteria, Virus = BaseSpecies(2)
    B1, B2 = New(Bacteria, 2)
    V1, V2 = New(Virus, 2)
    Bacteria.not_infected + Virus >> Bacteria.infected[1]
    MySim = Simulation(B1 | B2 | V1 | V2)
    MySim.level = -1
    results = MySim.compile()
    assert compare_model(results, 'test_tools/model_4.txt')


# Model to test round-robin and stoichiometry
def test_model_5():
    A = BaseSpecies(1)
    B, C = New(A, 2)
    A >> 2 * A[1]
    2 * A >> 3 * A[1]
    MySim = Simulation(B | C)
    MySim.level = -1
    results = MySim.compile()
    assert compare_model(results, 'test_tools/model_5.txt')


def test_model_6():
    # This model tests species that are not referenced in the reactants (we call them Born Species)
    A = BaseSpecies(1)
    B = New(A)
    C = New(A)
    B.b1, B.b2, C.c1, C.c2
    Zero >> 2 * A[1]
    MySim = Simulation(B | C)
    MySim.level = -1
    results = MySim.compile()
    assert compare_model(results, 'test_tools/model_6.txt')


def test_model_7():
    def oscillator(beta_m=5, beta_p=10, gamma_m=1, gamma_p=0.01, k=1, n=4, leaky=0.0001):
        Mortal, Creator = BaseSpecies(2)

        mRNA = Mortal * Creator
        Protein = New(Mortal)

        # Repression reactions
        for m, p in zip(['m1', 'm2', 'm3'], ['x2', 'x3', 'x1']):
            Protein.c(p) >> Protein.c(p) + mRNA.c(m)[lambda pro: f'{beta_m}/(1 + ({pro}/{k})^{n})']

        # Production reactions
        for m, p in zip(['m1', 'm2', 'm3'], ['x1', 'x2', 'x3']):
            mRNA.c(m) >> mRNA.c(m) + Protein.c(p)[beta_p]

        # We need the rate of degradation to be different from proteins and mRNA
        Mortal >> Zero[lambda r1: gamma_p if r1.is_a(Protein) else gamma_m]
        # This is the leaky mRNA expression, it needs to be low
        Zero >> Creator[leaky]

        MySim = Simulation(mRNA | Protein)
        MySim.level = -1
        return MySim.compile()

    assert compare_model(oscillator(), 'test_tools/model_7.txt')


# Model to test well defined orthogonal spaces
def test_orthogonal_spaces():
    try:
        A, B = BaseSpecies(2)
        A.a, A.b
        C = New(B)
        C.a, C.b
        MySim = Simulation(A | C)
        MySim.level = -1
        MySim.compile()
        assert False
    except SystemExit:
        assert True


# Model to test dimensional inconsistency
def test_dimensional_inconsistency():
    try:
        A, B, C = BaseSpecies(3)
        A(1 * u.mol / u.meter ** 3) + B(1 * u.mol / u.meter ** 2) >> C[1]
        MySim = Simulation(A | B | C)
        MySim.level = -1
        MySim.compile()
        assert False
    except SystemExit:
        print('Dimensional inconsistency model Ok')
        assert True


def test_average_value():
    E = BaseSpecies(1)
    Zero >> E[12]
    E >> Zero[25]

    MySim = Simulation(E)
    MySim.save_data = False
    MySim.plot_data = False
    MySim.level = -1
    MySim.run()


def test_hybrid_sim():
    A, B = BaseSpecies(2)
    A >> 2 * A[1]

    A(1)
    S1 = Simulation(A | B)
    S1.save_data = False
    S1.plot_data = False
    S1.duration = 3
    S1.level = -1

    A.reset_reactions()
    A + B >> Zero[0.01]

    B(50)
    S2 = Simulation(A | B)
    S2.method = 'stochastic'
    S2.duration = (A <= 0) | (B <= 0)
    S2.level = -1

    Sim = S1 + S2
    Sim.run()
    assert compare_model(Sim.compile(), 'test_tools/model_8.txt')
    assert Sim.results[A][-1] == 0 or Sim.results[B][-1] == 0


def test_concatenated_simulation():
    A, B, C = BaseSpecies(3)
    A >> Zero[1]

    A(50)
    S1 = Simulation(A)
    S1.plot_data = False
    S1.duration = 5
    S1.level = -1

    B >> Zero[1]

    B(50)
    S2 = Simulation(B)
    S2.duration = 5
    S2.level = -1

    C >> Zero[1]

    C(50)
    S3 = Simulation(C)
    S3.duration = 5
    S3.level = -1

    S = S1 + S2 + S3
    S.run()
    assert S.results[A][-1] < 1 and S.results[B][-1] < 1 and S.results[C][-1] < 1


def test_event_type():
    A, B, C, D, E, F = BaseSpecies(6)

    A + B >> Zero[1]

    A(50), B(50), C(0)
    S = Simulation(A | B | C | D | E | F)
    S.plot_data = False
    S.level = -1

    with S.event_time(0):
        F(1)

    with S.event_condition((A <= 1) & (B <= 1)):
        C(1)

    with S.event_condition((A <= 1) & (B <= 1)):
        D(1)

    with S.event_condition(B <= 1):
        E(1)

    S.duration = 5
    assert compare_model(S.compile(), 'test_tools/model_15.txt')


def test_reacting_species_event():
    B = BaseSpecies(1)
    B.b1, B.b2
    A = New(B)

    B >> Zero[1]
    A.b2 >> Zero[0.5]
    A.a1 >> Zero[1]

    A.a1(100), A.b2(100), B.b1(100)
    S = Simulation(A | B)
    S.level = -1

    with S.event_condition((A.a1 <= 10) & (B.b1 <= 10)):
        A.a1(100)

    S.duration = 5
    assert compare_model(S.compile(), 'test_tools/model_9.txt')


def test_unit_event_test():
    A = BaseSpecies(1)
    A >> Zero[1 / u.s]

    A(1 * u.mol)
    S = Simulation(A)
    S.level = -1
    with S.event_condition(A < 0.5 * u.mol):
        A(1 * u.mol)
    S.duration = 3
    assert compare_model(S.compile(), 'test_tools/model_10.txt')


def test_reaction_deactivation():
    A, R = BaseSpecies(2)
    A + R >> 2 * A + R[1]

    A(1), R(1)
    S1 = Simulation(A | R)
    S1.level = -1
    S1.duration = 1
    S1.plot_data = False

    R(0)
    S2 = Simulation(A | R)
    S2.duration = 1
    S2.level = -1

    Sim = S1 + S2
    Sim.run()
    assert Sim.results[A][0] < Sim.results[A][-1] and Sim.results[R][0] == 1 and Sim.results[R][-1] == 0


def test_count_assignment():
    A = BaseSpecies(1)
    B = New(A)
    A.a1, A.a2

    B.b1 >> Zero[1]

    A.a1(100), A.a2(100)
    B.b1(100), B.b2(100)
    S = Simulation(A | B)
    S.level = -1
    S.plot_data = False
    S.duration = 5
    S.run()
    assert compare_model(S.compile(), 'test_tools/model_11.txt') \
           and 150 > S.results[B][-1] > 100 and S.results[A][-1] == 200


def test_reaction_deactivation():
    A, R = BaseSpecies(2)
    A + R.r1 >> 2 * A + R.r1[1]

    A(1), R(1)
    S1 = Simulation(A | R)
    S1.plot_data = False
    S1.duration = 2
    S1.level = -1

    R(0)
    S2 = Simulation(A | R)
    S2.duration = 2
    S2.level = -1
    Sim = S1 + S2
    Sim.run()
    assert Sim.results[R][-1] == 0 and Sim.results[R][0] == 1 and Sim.results[A][-2] == Sim.results[A][-1]


def test_complex_cell_model():
    Resource, Phage, Infectable = BaseSpecies(3)
    Cell = New(Infectable)
    Cell.t1, Cell.t2, Cell.t3

    def reproduction_rate(r):
        if r.t1:
            return 2
        elif r.t2:
            return 1
        elif r.t3:
            return 0.5

    Infectable.not_infected + Phage >> Infectable.infected[1]
    Cell + Resource >> 2 * Cell[lambda r1: reproduction_rate(r1)]
    Zero >> Resource[1]
    Resource >> Zero[1]
    Cell >> Zero[0.1]

    Cell.t1(1), Cell.t2(1), Cell.t3(1)
    S1 = Simulation(Cell | Resource | Phage)
    S1.level = -1
    S1.duration = 30

    Cell.reset_reactions()
    Phage(1000)
    S2 = Simulation(Cell | Phage)
    S2.level = -1
    S2.duration = 10
    S = S1 + S2
    S.plot_data = False
    S.run()

    for i, c in enumerate(S.results['Cell.t1.not_infected']):
        if c > 0:
            continue
        else:
            change_index = i
            break

    boll_1 = round(S.results[Cell.t1.not_infected][change_index - 1], 2) \
             == round(S.results[Cell.t1.infected][-1], 2)
    boll_2 = round(S.results[Cell.t1.not_infected][change_index - 1], 2) \
             == round(S.results[Cell.t1.infected][-1], 2)
    boll_3 = round(S.results[Cell.t1.not_infected][change_index - 1], 2) \
             == round(S.results[Cell.t1.infected][-1], 2)
    assert boll_1 and boll_2 and boll_3


def test_zero_rate_reactions():
    A, B = BaseSpecies(2)

    A.a1, A.a2, B.b1, B.b2
    Combination = A * B

    Combination >> Zero[lambda r1: 0 if r1.b2 else 1]
    S = Simulation(Combination)
    S.level = -1
    return compare_model(S.compile(), 'test_tools/model_12.txt')


def test_double_rate():
    A, B = BaseSpecies(2)
    A.a1, A.a2, B.b1, B.b2

    def rate(r1, ball):
        factor1 = 0.5 if r1.a1 else 1
        factor2 = 0.5 if ball.b1 else 1
        return factor1 * factor2

    A + B >> Zero[rate]
    S = Simulation(A | B)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_13.txt')


def test_single_rate():
    A, B = BaseSpecies(2)
    A.a1, A.a2, B.b1, B.b2

    def rate(r1):
        factor1 = 0.5 if r1.a1 else 1
        return factor1

    A + B >> Zero[rate]
    S = Simulation(A | B)
    S.level = - 1
    assert compare_model(S.compile(), 'test_tools/model_14.txt')


def test_triple_rate():
    A, B = BaseSpecies(2)
    A.a1, A.a2, B.b1, B.b2

    def rate(r1, r2, r3):
        factor1 = 0.5 if r1.a1 else 1
        factor2 = 0.5 if r2.b1 else 1
        factor3 = 0.5 if r3.c1 else 1
        return factor1*factor2*factor3

    A + B >> Zero[rate]
    S = Simulation(A | B)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_13.txt')


def test_stochastic_event_duration():

    A, B = BaseSpecies(2)
    A + B >> Zero[0.01]

    A(100), B(100)
    S1 = Simulation(A | B)
    S1.save_data = False
    S1.plot_data = False
    S1.method = 'stochastic'
    S1.duration = (A <= 0) | (B <= 0)
    S1.level = -1
    S1.run()
    R = S1.results
    assert R[A][0] > 0 and R[B][0] > 0 and R[A][-1] == 0 and R[B][-1] == 0


def test_logic_operator_syntax():
    test_failed = False
    simlog.global_simlog_level = -1
    A, B = BaseSpecies(2)
    A.a1, A.a2, A.a3

    try:
        (10 >= A) >= 10
        test_failed = True
    except SystemExit:
        pass

    try:
        (10 >= A >= 10 >= A)
        test_failed = True
    except SystemExit:
        pass

    try:
        (10 >= A*A)
        test_failed = True
    except SystemExit:
        pass

    try:
        S1 = Simulation(A)
        S1.level = -1
        with S1.event_condition(B <= 10):
            A(100)
        S1.compile()
        test_failed = True
    except SystemExit:
        pass

    r1 = ((10 >= 2*A) & (A <= 10)) | (10 >= A)
    S = Simulation(A)
    S.level = -1
    with S.event_condition(r1):
        A(100)
    assert compare_model(S.compile(), 'test_tools/model_16.txt')

    if test_failed:
        assert False


def test_stack_position():

    Cell = BaseSpecies()
    Cell >> 2 * Cell[1]
    A, B, C = New(Cell)
    S = Simulation(A | B | C)
    S.level = -1
    compare_model(S.compile(), 'test_tools/model_17.txt')

    def hi():
        Cell = BaseSpecies()
        Cell >> 2 * Cell[1]
        A, B, C = New(Cell)

        S = Simulation(A | B | C)
        S.level = -1
        compare_model(S.compile(), 'test_tools/model_17.txt')
    hi()

    def hi_inside_hi():
        hi()
    hi_inside_hi()


def test_empty_arguments():
    A, B = BaseSpecies()

    A >> Zero[lambda: f'{A}*0.01']
    S = Simulation(A)
    S.duration = 5
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_18.txt')


def test_conditional_between_meta_species():
    Cu = BaseSpecies()
    Cu.c1, Cu.c2
    Azi, Byy = New(Cu)
    Azi.a1, Azi.a2, Byy.b1, Byy.b2

    Azi >> Zero[1]
    Byy >> Zero[0.1]

    Azi(200), Byy(50)
    S = Simulation(Azi | Byy)
    S.plot_data = False
    S.level = -1
    with S.event_condition(Azi.a1 <= Byy.b1):
        Azi(200)
    S.duration = 10
    S.method = 'stochastic'
    assert compare_model(S.compile(), 'test_tools/model_19.txt')
    S.run()
    for i in [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]:
        assert S.results[Azi][i] > S.results[Byy][i]


def test_conditional_between_meta_species_2():
    A, B = BaseSpecies()

    A >> Zero[1]
    B >> Zero[0.1]

    r1 = ((A < B) & (A < B) | (A < B))

    A(200), B(50)
    S = Simulation(A | B)
    with S.event_condition(r1):
        A(200)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_20.txt')


def test_event_reaction_not_allowed():

    try:
        A = BaseSpecies()
        A >> Zero[1]

        S = Simulation(A)

        with S.event_time(0):
            Zero >> A[1]
        assert False
    except SystemExit:
        assert True


def all_test():
    A, B = BaseSpecies()
    A.a1, A.a2
    B.b1, B.b2

    C = A * B

    All[C](100)
    C >> All[C][1]

    S = Simulation(C)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_21.txt')


def all_test_2():
    B = BaseSpecies()
    B.b1, B.b2
    C, D = New(B)
    C.c1, C.c2, D.d1, D.d2

    Zero >> All[B.b1][1]

    S = Simulation(C | D)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_22.txt')


def test_error_mult():
    try:
        D = BaseSpecies(1)
        A, B, C = D * BaseSpecies(3)
        simlog.global_simlog_level = -1
        assert False
    except SystemExit:
        assert True


def test_set_counts():
    A, C = BaseSpecies()
    A.a1, A.a2
    B = New(A)
    B.b1, B.b2

    model = set_counts({All['B.a1']: 100, C:200*u.mols, 'A.a1': 100, A.a2: 50})
    S = Simulation(model)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_23.txt')


def test_bool_error():
    B = BaseSpecies()

    B >> Zero[1]
    B(100)

    S = Simulation(B)

    simlog.global_simlog_level = -1
    try:
        with S.event_condition(B == 0):
            B(100)
        assert False
    except SystemExit:
        pass

    try:
        S.duration = True
        assert False
    except SystemExit:
        pass
    assert True


def test_event_all():
    Acka = BaseSpecies()
    Acka.a1, Acka.a2

    Baka = New(Acka)
    Baka.b1, Baka.b2

    Baka >> Zero[1]

    S = Simulation(Baka)
    S.level = -1
    S.plot_data = False
    with S.event_time(0):
        set_counts({All[Baka]: 10, Baka.b1.a2: 20})
        Baka.a1.b1(30)
    assert compare_model(S.compile(), 'test_tools/model_24.txt')
    S.duration = 5
    S.step_size = 1
    S.run()
    assert S.results[Baka.a1.b1][0] == 30
    assert S.results[Baka.b1.a2][0] == 20
    assert S.results[Baka.a1.b2][0] == 10
    assert S.results['Baka.a1.b1'][0] == 30
    assert S.results['Baka.b1.a2'][0] == 20
    assert S.results['Baka.a1.b2'][0] == 10
    for key in S.results:
        if key == 'Time':
            continue
        assert S.results[key][-1] < 1


def test_one_value_concatenation_sim():
    A, B = BaseSpecies()

    B(200)
    S2 = Simulation(A | B)
    S2.plot_data = False
    S2.level = -1
    S2.duration = 5
    S2.step_size = 1
    S2.duration = (A <= 0) | (B <= 0)
    S2.run()
    assert len(S2.results[A]) == 1


def test_crash_after_modification():

    try:
        A = BaseSpecies()
        S1 = Simulation(A)
        A = BaseSpecies()
        A.a1, A.a2

        S2 = Simulation(A)
        S = S1 + S2
        S.level = -1
        S.run()
        assert False
    except SystemExit:
        assert True


def test_unit_bi_dimension():
    A = BaseSpecies()
    A(5 / u.m ** 2)
    S = Simulation(A)
    S.volume = 2 * u.m ** 2
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_25.txt')


def test_bi_dimensional_rates():
    Ball, Child, Bacteria = BaseSpecies(3)

    Ball(10 / u.meter ** 2)
    Child(1 / u.meter ** 2)
    Bacteria(1 * u.mol)

    Bacteria >> Zero[1 * u.mol / u.second]
    Ball + Child + Child >> Ball + Child[1e-3 * (u.meter ** 4) / u.hour]
    Ball + Child >> Ball[1e-3 * (u.meter ** 2) / u.hour]

    S = Simulation(Ball | Child | Bacteria)
    S.volume = 2 * u.m ** 2
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_26.txt')


def test_dimension_in_function_only():
    A = BaseSpecies()

    A + A >> 3 * A[lambda: 1 * u.milliliter / u.second]

    A(1)
    S = Simulation(A)
    S.level = -1
    assert compare_model(S.compile(), 'test_tools/model_27.txt')


test_list = [test_model_1, test_model_2, test_model_3, test_model_4, test_model_5, test_model_6, test_model_7,
             test_orthogonal_spaces, test_average_value, test_hybrid_sim, test_concatenated_simulation,
             test_event_type, test_reacting_species_event, test_unit_event_test, test_reaction_deactivation,
             test_double_rate, test_single_rate, test_triple_rate, test_stochastic_event_duration,
             test_logic_operator_syntax, test_stack_position, test_empty_arguments,
             test_conditional_between_meta_species, test_conditional_between_meta_species_2,
             test_event_reaction_not_allowed, all_test, all_test_2, test_error_mult, test_set_counts,
             test_bool_error, test_event_all, test_one_value_concatenation_sim, test_crash_after_modification,
             test_unit_bi_dimension, test_bi_dimensional_rates, test_dimension_in_function_only]

sub_test = test_list


def perform_tests():

    any_failed = False
    for test in sub_test:
        try:
            test()
            print(f'Test {test} passed')
        except:
            print('\033[91m' + f'Test {test} failed'  + '\033[0m', file=sys.stderr)
            any_failed = True

    if any_failed:
        assert False


perform_tests()
